"""These tests apply to only samplable distributions."""
from __future__ import annotations

from typing import Any

import jax.numpy as jnp
from jax import jacrev, vmap
from jax.random import split
from numpy.random import Generator
from tjax import JaxArray, JaxRealArray, KeyArray, hessian

from efax import Flattener, Samplable, SimpleDistribution, flatten_mapping

from .distribution_info import DistributionInfo


def sample_using_flattened(flattened_parameters: JaxRealArray,
                           flattener: Flattener[Any],
                           key: KeyArray,
                           ) -> JaxArray:
    p = flattener.unflatten(flattened_parameters)

    def flattened_sample(q: SimpleDistribution) -> JaxRealArray:
        assert isinstance(q, Samplable)
        ordinary_sample = q.sample(key)
        support = q.domain_support()
        return support.flattened(ordinary_sample, map_to_plane=False)

    if isinstance(p, SimpleDistribution):
        return flattened_sample(p)
    sample_dict = p.general_method(flattened_sample)
    sample_flat = flatten_mapping(sample_dict)
    return jnp.concat(tuple(sample_flat.values()), axis=-1)


def calculate_jacobian(p: SimpleDistribution,
                       keys: KeyArray,
                       ) -> JaxRealArray:
    flattener, flattened = Flattener.flatten(p)
    jacobian_sample = vmap(jacrev(sample_using_flattened, argnums=(0,)), in_axes=(0, 0, 0))
    parameters_jacobian, = jacobian_sample(flattened, flattener, keys)
    assert parameters_jacobian.shape[0: 1] == p.shape
    parameters_jacobian = jnp.sum(parameters_jacobian, axis=0)
    return jnp.sum(parameters_jacobian, axis=0)


def calculate_curvature(p: SimpleDistribution,
                        keys: KeyArray,
                        ) -> JaxRealArray:
    # Calculate curvature.
    flattener, flattened = Flattener.flatten(p)
    hessian_sample = vmap(hessian(sample_using_flattened, argnums=(0,), reverse_only=True),
                          in_axes=(0, 0, 0))
    parameters_hessian_x, = hessian_sample(flattened, flattener, keys)
    parameters_hessian, = parameters_hessian_x

    assert parameters_hessian.shape[0: 1] == p.shape
    parameters_hessian = jnp.sum(parameters_hessian, axis=0)

    assert parameters_hessian.shape[-2] == parameters_hessian.shape[-1]
    diagonal_hessian = jnp.linalg.diagonal(parameters_hessian)
    # There's one element for each sample element.  Just sum over the curvatures.
    return jnp.sum(diagonal_hessian, axis=0)


def test_sampling_cotangents(generator: Generator,
                             key: KeyArray,
                             sampling_wc_distribution_info: DistributionInfo[Any, Any, Any],
                             *,
                             distribution_name: str | None,
                             natural: bool) -> None:
    """Test that the curvature is nonzero."""
    info = sampling_wc_distribution_info
    info.skip_if_deselected(distribution_name)
    distribution_shape = (23,)
    keys = split(key, distribution_shape)

    p: SimpleDistribution = (info.nat_parameter_generator(generator, distribution_shape)
                          if natural
                          else info.exp_parameter_generator(generator, distribution_shape))
    jacobian = calculate_jacobian(p, keys)
    curvature = calculate_curvature(p, keys)
    assert jacobian.shape == curvature.shape
    assert jnp.all((jacobian != 0) | (curvature != 0))



from collections.abc import Callable, Mapping
from functools import reduce
from typing import Any, override

from jax.random import split
from tjax import JaxComplexArray, JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parametrization import Distribution, SimpleDistribution
from ..tools import join_mappings


@dataclass
class JointDistribution(Distribution):
    _sub_distributions: Mapping[str, Distribution]

    @override
    def sub_distributions(self) -> Mapping[str, Distribution]:
        return self._sub_distributions

    def general_method(self,
                       f: Callable[[Distribution], Any],
                       t: type[Distribution] = Distribution
                       ) -> Any:
        return {name: (value.general_method(f, t)
                       if isinstance(value, JointDistribution) else f(value))
                for name, value in self._sub_distributions.items()
                if isinstance(value, JointDistribution | t)}

    def general_sample(self, key: KeyArray, shape: Shape | None = None) -> dict[str, Any]:
        count = 0

        def g(x: Distribution, /) -> None:
            nonlocal count
            count += 1

        self.general_method(g, Samplable)
        keys = split(key, count)
        count = 0

        def f(x: Distribution, /) -> JaxComplexArray:
            assert isinstance(x, Samplable)
            nonlocal keys, count
            retval = x.sample(keys[count], shape)
            count += 1
            return retval

        return self.general_method(f, Samplable)

    def as_dict(self) -> dict[str, Any]:
        def f(x: Distribution, /) -> SimpleDistribution:
            assert isinstance(x, SimpleDistribution)
            return x

        return self.general_method(f)

    @property
    @override
    def shape(self) -> Shape:
        for distribution in self._sub_distributions.values():
            return distribution.shape
        raise ValueError


@dataclass
class JointDistributionE(JointDistribution,
                         HasEntropyEP['JointDistributionN'],
                         ExpectationParametrization['JointDistributionN']):
    _sub_distributions: Mapping[str, ExpectationParametrization[Any]]

    @override
    def sub_distributions(self) -> Mapping[str, ExpectationParametrization[Any]]:
        return self._sub_distributions

    @override
    @classmethod
    def natural_parametrization_cls(cls) -> type['JointDistributionN']:
        return JointDistributionN

    @override
    def to_nat(self) -> 'JointDistributionN':
        return JointDistributionN({name: value.to_nat()
                                   for name, value in self._sub_distributions.items()})

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()

        def f(x: ExpectationParametrization[Any]) -> JaxRealArray:
            assert isinstance(x, HasEntropyEP)
            return x.expected_carrier_measure()

        return reduce(xp.add, (f(value) for value in self._sub_distributions.values()))


@dataclass
class JointDistributionN(JointDistribution,
                         HasEntropyNP[JointDistributionE],
                         NaturalParametrization[JointDistributionE, dict[str, Any]]):
    _sub_distributions: Mapping[str, NaturalParametrization[Any, Any]]

    @override
    def sub_distributions(self) -> Mapping[str, NaturalParametrization[Any, Any]]:
        return self._sub_distributions

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()

        return reduce(xp.add,
                      (x.log_normalizer() for x in self._sub_distributions.values()))

    @override
    def to_exp(self) -> JointDistributionE:
        return JointDistributionE({name: value.to_exp()
                                   for name, value in self._sub_distributions.items()})

    @override
    def carrier_measure(self, x: dict[str, Any]) -> JaxRealArray:
        xp = self.get_namespace()

        joined = join_mappings(sub=self._sub_distributions, x=x)
        return reduce(xp.add,
                      (value['sub'].carrier_measure(value['x'])
                       for value in joined.values()))

    @override
    @classmethod
    def sufficient_statistics(cls, x: dict[str, Any], **fixed_parameters: Any
                              ) -> JointDistributionE:
        raise NotImplementedError



from __future__ import annotations

import math
from typing import Any, cast

from tjax import JaxRealArray, Shape, inverse_softplus, softplus
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.multidimensional import Multidimensional
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import CircularBoundedSupport, VectorSupport, distribution_parameter
from ..tools import iv_ratio, log_ive


@dataclass
class VonMisesFisherNP(HasEntropyNP['VonMisesFisherEP'],
                       NaturalParametrization['VonMisesFisherEP', JaxRealArray],
                       Multidimensional):
    """The natural parametrization of the von Mises-Fisher distribution.

    Args:
        mean_times_concentration: E(x) times the concentration kappa.
    """
    mean_times_concentration: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_concentration.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        half_k = self.dimensions() * 0.5
        kappa = xp.linalg.vector_norm(self.mean_times_concentration, axis=-1)
        return (kappa
                - (half_k - 1.0) * xp.log(kappa)
                + half_k * math.log(2.0 * math.pi)
                + log_ive(half_k - 1.0, kappa))

    @override
    def to_exp(self) -> VonMisesFisherEP:
        xp = self.get_namespace()
        q = self.mean_times_concentration
        kappa: JaxRealArray = xp.linalg.vector_norm(q, axis=-1, keepdims=True)
        return VonMisesFisherEP(
                xp.where(kappa == 0.0,
                         q,
                         q * (_a_k(self.dimensions(), kappa) / kappa)))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(self.shape)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> VonMisesFisherEP:
        return VonMisesFisherEP(x)

    @override
    def dimensions(self) -> int:
        return self.mean_times_concentration.shape[-1]

    def kappa(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.linalg.vector_norm(self.mean_times_concentration, axis=-1)

    def to_kappa_angle(self) -> tuple[JaxRealArray, JaxRealArray]:
        if self.dimensions() != 2:  # noqa: PLR2004
            raise ValueError
        xp = self.get_namespace()
        kappa = self.kappa()
        angle = xp.where(kappa == 0.0,
                         xp.asarray(0.0),
                         xp.atan2(self.mean_times_concentration[..., 1],
                                  self.mean_times_concentration[..., 0]))
        return kappa, angle


@dataclass
class VonMisesFisherEP(HasEntropyEP[VonMisesFisherNP],
                       ExpToNat[VonMisesFisherNP],
                       Multidimensional,
                       ExpectationParametrization[VonMisesFisherNP]):
    """The expectation parametrization of the von Mises-Fisher distribution.

    Args:
        mean: E(x).
    """
    mean: JaxRealArray = distribution_parameter(CircularBoundedSupport(1.0))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[VonMisesFisherNP]:
        return VonMisesFisherNP

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def initial_search_parameters(self) -> JaxRealArray:
        xp = self.get_namespace()
        mu: JaxRealArray = xp.linalg.vector_norm(self.mean, axis=-1)
        # 0 <= mu <= 1.0
        initial_kappa = xp.where(mu == 1.0,
                                 xp.asarray(xp.inf),
                                 (mu * self.dimensions() - mu ** 3) / (1.0 - mu ** 2))
        initial_kappa = cast('JaxRealArray', initial_kappa)
        return inverse_softplus(initial_kappa)[..., xp.newaxis]

    @override
    def search_to_natural(self, search_parameters: JaxRealArray) -> VonMisesFisherNP:
        xp = self.get_namespace()
        kappa = softplus(search_parameters)
        mu = xp.linalg.vector_norm(self.mean, axis=-1, keepdims=True)
        q = self.mean * (kappa / mu)
        return VonMisesFisherNP(q)

    @override
    def search_gradient(self, search_parameters: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace()
        kappa = softplus(search_parameters)
        mu = xp.linalg.vector_norm(self.mean, axis=-1)
        return _a_k(self.dimensions(), kappa) - mu

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]


# Private functions --------------------------------------------------------------------------------
def _a_k(k: float | JaxRealArray, kappa: float | JaxRealArray) -> JaxRealArray:
    return iv_ratio(k * 0.5, kappa)



from __future__ import annotations

from typing import Any

import jax
from array_api_compat import get_namespace
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, ScalarSupport
from .dirichlet_common import DirichletCommonEP, DirichletCommonNP


@dataclass
class BetaNP(DirichletCommonNP['BetaEP'],
             NaturalParametrization['BetaEP', JaxRealArray]
             ):
    """The natural parametrization of the beta distribution.

    The best way to interpret the parameters of the beta distribution are that an observation x in
    [0, 1] represents the Bernoulli probability that outcome 0 (out of {0, 1}) is realized.  In this
    way, the Beta class coincides with a special case of the Dirichlet class.

    Args:
        alpha_minus_one: The shape parameters alpha, beta in an array with final dimension 2.
    """
    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=RealField(minimum=0.0, maximum=1.0))

    @override
    def to_exp(self) -> BetaEP:
        return BetaEP(self._exp_helper())

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> BetaEP:
        xp = get_namespace(x)
        return BetaEP(xp.stack([xp.log(x), xp.log1p(-x)], axis=-1))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.zeros(x.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        alpha = 1.0 + self.alpha_minus_one
        return jax.random.beta(key, alpha[..., 0], alpha[..., 1], shape)


@dataclass
class BetaEP(DirichletCommonEP[BetaNP],
             ExpectationParametrization[BetaNP]):
    """The expectation parametrization of the beta distribution.

    The best way to interpret the parameters of the beta distribution are that an observation x in
    [0, 1] represents the Bernoulli probability that outcome 0 (out of {0, 1}) is realized.  In this
    way, the Beta class coincides with a special case of the Dirichlet class.

    Args:
        mean_log_probability: An array with final dimension [E(log(x)), E(log(1-x))].
    """
    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=RealField(minimum=0.0, maximum=1.0))

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[BetaNP]:
        return BetaNP



from __future__ import annotations

from typing import Any

import numpy as np
from numpy.linalg import det, inv
from numpy.random import Generator
from numpy.testing import assert_allclose
from scipy.special import digamma, gammaln
from tjax import JaxRealArray

from efax import (ExpectationParametrization, GammaEP, GammaNP, MultivariateNormalEP,
                  MultivariateNormalNP, NaturalParametrization, NormalEP, NormalNP)

from .create_info import GammaInfo, MultivariateNormalInfo, NormalInfo
from .distribution_info import DistributionInfo


def prelude(generator: Generator,
            distribution_info_kl: DistributionInfo[Any, Any, Any],
            distribution_name: str | None
            ) -> tuple[ExpectationParametrization[Any], NaturalParametrization[Any, Any],
                       JaxRealArray]:
    shape = (3, 2)
    distribution_info_kl.skip_if_deselected(distribution_name)
    x = distribution_info_kl.exp_parameter_generator(generator, shape=shape)
    y = distribution_info_kl.nat_parameter_generator(generator, shape=shape)
    my_kl = x.kl_divergence(y)
    return x, y, my_kl


def test_normal_kl(generator: Generator, distribution_name: str | None) -> None:
    """Test the KL divergence."""
    x, y, my_kl = prelude(generator, NormalInfo(), distribution_name)
    assert isinstance(x, NormalEP)
    assert isinstance(y, NormalNP)
    xm = x.mean
    ym = y.to_exp().mean
    xv = x.variance()
    yv = y.to_exp().variance()
    scipy_kl = 0.5 * (xv / yv + np.square(ym - xm) / yv - 1.0 + np.log(yv / xv))
    rtol = 1e-5
    assert_allclose(my_kl, scipy_kl, rtol=rtol)


def test_mvn_kl(generator: Generator, distribution_name: str | None) -> None:
    """Test the KL divergence."""
    x, y, my_kl = prelude(generator, MultivariateNormalInfo(dimensions=4), distribution_name)
    assert isinstance(x, MultivariateNormalEP)
    assert isinstance(y, MultivariateNormalNP)
    xm = x.mean
    ym = y.to_exp().mean
    xv = x.variance()
    yv = y.to_exp().variance()
    md = ym - xm
    tq = np.trace(inv(yv) @ xv, axis1=-2, axis2=-1)
    mvm = np.einsum("...i,...ij,...j->...", md, inv(yv), md)
    scipy_kl = 0.5 * (tq + mvm - x.dimensions()
                      + np.log(det(yv) / det(xv)))
    rtol = 1e-5
    assert_allclose(my_kl, scipy_kl, rtol=rtol)


def test_gamma_kl(generator: Generator, distribution_name: str | None) -> None:
    """Test the KL divergence."""
    x, y, my_kl = prelude(generator, GammaInfo(), distribution_name)
    assert isinstance(x, GammaEP)
    assert isinstance(y, GammaNP)
    xsh = x.to_nat().shape_minus_one + 1.0
    ysh = y.shape_minus_one + 1.0
    xsc = -1.0 / x.to_nat().negative_rate
    ysc = -1.0 / y.negative_rate
    scipy_kl = (((xsc - ysc) / ysc) * xsh + gammaln(ysh) - gammaln(xsh)
                + ysh * np.log(ysc) - xsh * np.log(xsc)
                + (xsh - ysh) * (np.log(xsc) + digamma(xsh)))
    rtol = 1e-5
    assert_allclose(my_kl, scipy_kl, rtol=rtol)



from __future__ import annotations

from abc import abstractmethod
from dataclasses import dataclass
from math import comb, sqrt
from typing import Any, cast

import numpy as np
from array_api_compat import get_namespace
from jax.dtypes import canonicalize_dtype
from jax.scipy import special as jss
from numpy.random import Generator
from tjax import JaxArray, JaxComplexArray, JaxRealArray, Shape, inverse_softplus, softplus
from tjax.dataclasses import field
from typing_extensions import override

from .array_tools import array_at
from .types import Namespace


class Ring:
    @abstractmethod
    def num_elements(self, support_num_element: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        raise NotImplementedError

    @abstractmethod
    def unflattened(self, y: JaxRealArray, *, map_from_plane: bool) -> JaxArray:
        raise NotImplementedError

    @abstractmethod
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float) -> JaxRealArray:
        raise NotImplementedError

    def clamp(self, x: JaxArray) -> JaxArray:
        return x


@dataclass
class RealField(Ring):
    minimum: float | None = None
    maximum: float | None = None
    generation_scale: float = 1.0  # Scale the generated random numbers to improve random testing.
    min_open: bool = True  # Open interval
    max_open: bool = True  # Open interval

    def __post_init__(self) -> None:
        dtype = canonicalize_dtype(float)
        eps = float(np.finfo(dtype).eps)
        if self.min_open and self.minimum is not None:
            self.minimum = float(max(self.minimum + eps,
                                     self.minimum * (1.0 + np.copysign(eps, self.minimum))))
        if self.max_open and self.maximum is not None:
            self.maximum = float(min(self.maximum - eps,
                                     self.maximum * (1.0 + np.copysign(eps, -self.maximum))))

    @override
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        if not map_to_plane:
            return x
        match self.minimum, self.maximum:
            case None, float(maximum):
                return -inverse_softplus(maximum - x)
            case float(minimum), None:
                return inverse_softplus(x - minimum)
            case float(minimum), float(maximum):
                return jss.logit((x - minimum) / (maximum - minimum))
            case _:
                return x

    @override
    def unflattened(self, y: JaxRealArray, *, map_from_plane: bool) -> JaxArray:
        if not map_from_plane:
            return y
        match self.minimum, self.maximum:
            case None, float(maximum):
                return maximum - softplus(-y)
            case float(minimum), None:
                return softplus(y) + minimum
            case float(minimum), float(maximum):
                return minimum + jss.expit(y) * (maximum - minimum)
            case _:
                return y

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float) -> JaxRealArray:
        match self.minimum, self.maximum:
            case None, float(maximum):
                return xp.asarray(maximum - rng.exponential(size=shape) * self.generation_scale
                                  - safety)
            case float(minimum), None:
                return xp.asarray(rng.exponential(size=shape) * self.generation_scale + minimum
                                  + safety)
            case float(minimum), float(maximum):
                domain_width = xp.maximum(maximum - minimum - safety, 0.0)
                true_minimum = xp.mean(minimum, maximum) - domain_width * 0.5
                return xp.asarray(true_minimum + domain_width * rng.uniform(size=shape))
            case None, None:
                return xp.asarray(rng.normal(scale=1.0, size=shape) * self.generation_scale)
            case _:
                raise TypeError

    @override
    def clamp(self, x: JaxRealArray) -> JaxRealArray:
        xp = get_namespace(x)
        eps = xp.finfo(x.dtype).eps
        minimum = None if self.minimum is None else self.minimum + eps
        maximum = None if self.maximum is None else self.maximum - eps
        return xp.clip(x, min=minimum, max=maximum)


@dataclass
class ComplexField(Ring):
    minimum_modulus: float | JaxRealArray = 0.0
    maximum_modulus: float | JaxRealArray | None = None

    @override
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element * 2

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        xp = get_namespace(x)
        # if self.maximum_modulus is not None:
        #     assert False
        minimum = fix_bound(self.minimum_modulus, x)
        maximum = fix_bound(self.maximum_modulus, x)
        if map_to_plane:
            if maximum is None:
                # x is outside the disk of the given minimum.  Map it to the plane.
                magnitude = xp.abs(x)
                corrected_magnitude = magnitude - minimum
                corrected_x = x * corrected_magnitude / magnitude
            else:
                # x is in the disk of the given maximum.  Map it to the plane.
                magnitude = xp.abs(x)
                corrected_magnitude = jss.logit(((magnitude - minimum) / maximum) * 0.5 + 0.5)
                corrected_x = x * corrected_magnitude / magnitude
        else:
            corrected_x = x
        return xp.concat([xp.real(corrected_x), xp.imag(corrected_x)], axis=-1)

    @override
    def unflattened(self, y: JaxRealArray, *, map_from_plane: bool) -> JaxArray:
        assert y.shape[-1] % 2 == 0
        n = y.shape[-1] // 2
        corrected_x = y[..., :n] + 1j * y[..., n:]
        if not map_from_plane:
            return corrected_x
        return self._unflattened(corrected_x)

    def _unflattened(self, corrected_x: JaxComplexArray) -> JaxComplexArray:
        xp = get_namespace(corrected_x)
        minimum = fix_bound(self.minimum_modulus, corrected_x)
        maximum = fix_bound(self.maximum_modulus, corrected_x)

        if maximum is None:
            # x is outside the disk of the given minimum.  Map it to the plane.
            corrected_magnitude = xp.abs(corrected_x)
            magnitude = corrected_magnitude + minimum
            return corrected_x * magnitude / corrected_magnitude
        # x is in the disk of the given maximum.  Map it to the plane.
        corrected_magnitude = xp.abs(corrected_x)
        magnitude = maximum * (jss.expit(corrected_magnitude) - 0.5) * 2.0 + minimum
        return corrected_x * magnitude / corrected_magnitude

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float) -> JaxRealArray:
        return self._unflattened(xp.asarray(rng.normal(size=shape) + 1j * rng.normal(size=shape)))


class BooleanRing(Ring):
    @override
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.asarray(x, dtype=canonicalize_dtype(float))

    @override
    def unflattened(self, y: JaxRealArray, *, map_from_plane: bool) -> JaxArray:
        xp = get_namespace(y)
        return xp.asarray(y, dtype=xp.bool_)

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float) -> JaxRealArray:
        return xp.asarray(rng.binomial(1, 0.5, shape), dtype=xp.bool_)


@dataclass
class IntegralRing(Ring):
    minimum: int | None = None
    maximum: int | None = None

    @override
    def num_elements(self, support_num_element: int) -> int:
        return support_num_element

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.asarray(x, dtype=canonicalize_dtype(float))

    @override
    def unflattened(self, y: JaxRealArray, *, map_from_plane: bool) -> JaxArray:
        xp = get_namespace(y)
        return xp.asarray(y, dtype=canonicalize_dtype(int))

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float) -> JaxRealArray:
        field = RealField(minimum=None if self.minimum is None else float(self.minimum),
                      maximum=None if self.maximum is None else float(self.maximum))
        real_values = field.generate(xp, rng, shape, safety) * 10
        return xp.astype(real_values, canonicalize_dtype(int))


real_field = RealField()
negative_support = RealField(maximum=0.0)
positive_support = RealField(minimum=0.0)
integral_ring = IntegralRing()
complex_field = ComplexField()
boolean_ring = BooleanRing()


class Support:
    @override
    def __init__(self, *, ring: Ring = real_field) -> None:
        super().__init__()
        self.ring = ring

    @abstractmethod
    def axes(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def shape(self, dimensions: int) -> Shape:
        raise NotImplementedError

    @abstractmethod
    def num_elements(self, dimensions: int) -> int:
        raise NotImplementedError

    @abstractmethod
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        raise NotImplementedError

    @abstractmethod
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        raise NotImplementedError

    @abstractmethod
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float, dimensions: int
                 ) -> JaxRealArray:
        raise NotImplementedError

    def clamp(self, x: JaxArray) -> JaxArray:
        return self.ring.clamp(x)


class ScalarSupport(Support):
    @override
    def axes(self) -> int:
        return 0

    @override
    def shape(self, dimensions: int) -> Shape:
        return ()

    @override
    def num_elements(self, dimensions: int) -> int:
        return self.ring.num_elements(1)

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        xp = get_namespace(x)
        return self.ring.flattened(x[..., xp.newaxis], map_to_plane=map_to_plane)

    @override
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        x = self.ring.unflattened(y, map_from_plane=map_from_plane)
        assert x.shape[-1] == 1
        return x[..., 0]

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float, dimensions: int
                 ) -> JaxRealArray:
        return self.ring.generate(xp, rng, shape, safety)


class VectorSupport(Support):
    @override
    def axes(self) -> int:
        return 1

    @override
    def shape(self, dimensions: int) -> Shape:
        return (dimensions,)

    @override
    def num_elements(self, dimensions: int) -> int:
        return self.ring.num_elements(dimensions)

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        return self.ring.flattened(x, map_to_plane=map_to_plane)

    @override
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        x = self.ring.unflattened(y, map_from_plane=map_from_plane)
        assert x.shape[-1] == dimensions
        return x

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float, dimensions: int
                 ) -> JaxRealArray:
        return self.ring.generate(xp, rng, (*shape, dimensions), safety)


class SimplexSupport(Support):
    @override
    def axes(self) -> int:
        return 1

    @override
    def shape(self, dimensions: int) -> Shape:
        return (dimensions - 1,)

    @override
    def num_elements(self, dimensions: int) -> int:
        return self.ring.num_elements(dimensions - 1)

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        return self.ring.flattened(x, map_to_plane=map_to_plane)

    @override
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        x = self.ring.unflattened(y, map_from_plane=map_from_plane)
        assert x.shape[-1] == dimensions
        return x

    @override
    def clamp(self, x: JaxArray) -> JaxArray:
        xp = get_namespace(x)
        eps = xp.finfo(x.dtype).eps
        s = xp.sum(x, axis=-1, keepdims=True)
        x *= xp.minimum(1.0, (1.0 - eps) / s)
        return xp.clip(x, min=eps, max=1.0 - eps)

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float, dimensions: int
                 ) -> JaxRealArray:
        raise NotImplementedError


class SymmetricMatrixSupport(Support):
    @override
    def __init__(self,
                 *,
                 positive_semidefinite: bool = False,
                 negative_semidefinite: bool = False,
                 hermitian: bool = False,
                 **kwargs: Any) -> None:
        if hermitian:
            kwargs.setdefault('ring', complex_field)
        super().__init__(**kwargs)
        self.hermitian = hermitian
        self.positive_semidefinite = positive_semidefinite
        self.negative_semidefinite = negative_semidefinite

    @override
    def axes(self) -> int:
        return 2

    @override
    def shape(self, dimensions: int) -> Shape:
        return (dimensions, dimensions)

    @override
    def num_elements(self, dimensions: int) -> int:
        return self.ring.num_elements(comb(dimensions + 1, 2))

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        xp = get_namespace(x)
        dimensions = x.shape[-1]
        assert x.shape[-2] == dimensions
        index_a, index_b = np.triu_indices(dimensions)
        x_triangular = xp.stack([x[..., i, j]
                                 for i, j in zip(index_a, index_b, strict=True)],
                                axis=-1)
        return self.ring.flattened(x_triangular, map_to_plane=map_to_plane)

    @override
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        xp = get_namespace(y)
        x = self.ring.unflattened(y, map_from_plane=map_from_plane)
        k = x.shape[-1]
        sqrt_discriminant = sqrt(1 + 8 * k)
        i_sqrt_discriminant = int(sqrt_discriminant)
        if i_sqrt_discriminant != sqrt_discriminant:
            msg = f"{k} {sqrt_discriminant}"
            raise ValueError(msg)
        if i_sqrt_discriminant % 2 != 1:
            raise ValueError
        dimensions = (i_sqrt_discriminant - 1) // 2
        index_a, index_b = np.triu_indices(dimensions)
        result = xp.empty(x.shape[:-1] + (dimensions, dimensions), dtype=x.dtype)
        for k, (i, j) in enumerate(zip(index_a, index_b, strict=True)):
            xk = x[..., k]
            result = array_at(result)[..., i, j].set(xk)
            if i != j:
                cxk = xp.conj(xk) if self.hermitian else xk
                result = array_at(result)[..., j, i].set(cxk)
        return result

    @override
    def generate(self, xp: Namespace, rng: Generator, shape: Shape, safety: float, dimensions: int
                 ) -> JaxRealArray:
        # Generate a random matrix.
        m = self.ring.generate(xp, rng, (*shape, dimensions, dimensions), safety)
        if self.negative_semidefinite or self.positive_semidefinite:
            # Perform QR decomposition to obtain an orthogonal matrix Q
            q, _ = np.linalg.qr(m)
            # Generate Eigenvalues.
            assert isinstance(self.ring, RealField | ComplexField)
            eig_field = (RealField(minimum=0.0)
                         if self.positive_semidefinite else RealField(maximum=0.0))
            eig = eig_field.generate(xp, rng, (*shape, dimensions), safety)
            # Return Q.T @ diag(eig) @ Q.
            return xp.einsum('...ji,...j,...jk->...ik', xp.conj(q) if self.hermitian else q, eig, q)
        mt = xp.matrix_transpose(m)
        if self.hermitian:
            mt = xp.conj(mt)
        return m + mt


class SquareMatrixSupport(Support):
    @override
    def axes(self) -> int:
        return 2

    @override
    def shape(self, dimensions: int) -> Shape:
        return (dimensions, dimensions)

    @override
    def num_elements(self, dimensions: int) -> int:
        return self.ring.num_elements(dimensions ** 2)

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxArray:
        xp = get_namespace(x)
        y = xp.reshape(x, (*x.shape[:-2], -1))
        return self.ring.flattened(y, map_to_plane=map_to_plane)

    @override
    def unflattened(self, y: JaxArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        xp = get_namespace(y)
        x = self.ring.unflattened(y, map_from_plane=map_from_plane)
        return xp.reshape(x, x.shape[:-1] + self.shape(dimensions))


class CircularBoundedSupport(VectorSupport):
    def __init__(self, radius: float) -> None:
        super().__init__()
        self.radius = radius

    @override
    def flattened(self, x: JaxArray, *, map_to_plane: bool) -> JaxRealArray:
        if not map_to_plane:
            return x
        # x is in the disk of the given radius.  Map it to the plane.
        xp = get_namespace(x)
        magnitude = xp.linalg.norm(x, 2, axis=-1, keepdims=True)
        corrected_magnitude = jss.logit((magnitude / self.radius) * 0.5 + 0.5)
        return x * corrected_magnitude / magnitude

    @override
    def unflattened(self, y: JaxRealArray, dimensions: int, *, map_from_plane: bool) -> JaxArray:
        if not map_from_plane:
            return y
        # y is in the plane.  Map it to the disk of the given radius.
        xp = get_namespace(y)
        assert y.shape[-1] == dimensions
        corrected_magnitude = cast('JaxRealArray', xp.linalg.norm(y, 2, axis=-1, keepdims=True))
        magnitude = self.radius * (jss.expit(corrected_magnitude) - 0.5) * 2.0
        return y * magnitude / corrected_magnitude


def distribution_parameter(support: Support,
                           *,
                           fixed: bool = False,
                           static: bool = False) -> Any:
    if static and not fixed:
        raise ValueError
    return field(static=static, metadata={'support': support, 'fixed': fixed, 'parameter': True})


def fix_bound(x: JaxArray | float | None, y: JaxArray) -> JaxArray | None:
    xp = get_namespace(y)
    if x is None:
        return x
    if isinstance(x, float):
        x = xp.asarray(x)
    assert isinstance(x, JaxArray)
    xp = get_namespace(x, y)
    while x.ndim < y.ndim:
        x = x[..., xp.newaxis]
    return x



from __future__ import annotations

from abc import abstractmethod

from ..parametrization import SimpleDistribution


class Multidimensional(SimpleDistribution):
    @abstractmethod
    def dimensions(self) -> int:
        raise NotImplementedError



from __future__ import annotations

from typing import Any

import jax
import numpy as np
from array_api_compat import get_namespace
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import ScalarSupport, distribution_parameter, negative_support, positive_support
from ..parametrization import SimpleDistribution


@dataclass
class WeibullNP(HasEntropyNP['WeibullEP'],
                Samplable,
                NaturalParametrization['WeibullEP', JaxRealArray],
                SimpleDistribution):
    """The natural parametrization of the Weibull distribution.

    Args:
        concentration: The fixed shape parameter k.
        eta: -scale ^ -concentration.
    """
    concentration: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support),
                                                         fixed=True)
    eta: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.eta.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return -xp.log(-self.eta) - xp.log(self.concentration)

    @override
    def to_exp(self) -> WeibullEP:
        return WeibullEP(self.concentration, -1.0 / self.eta)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return (self.concentration - 1.0) * xp.log(x)

    @override
    @classmethod
    def sufficient_statistics(cls,
                              x: JaxRealArray,
                              **fixed_parameters: Any
                              ) -> WeibullEP:
        xp = get_namespace(x)
        concentration = fixed_parameters['concentration']
        return WeibullEP(xp.broadcast_to(concentration, x.shape), x ** concentration)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class WeibullEP(HasEntropyEP[WeibullNP],
                ExpectationParametrization[WeibullNP],
                Samplable,
                SimpleDistribution):
    """The expectation parametrization of the Weibull distribution.

    Args:
        concentration: The fixed shape parameter k.
        chi: scale ^ concentration.
    """
    concentration: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support),
                                                         fixed=True)
    chi: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.chi.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[WeibullNP]:
        return WeibullNP

    @override
    def to_nat(self) -> WeibullNP:
        return WeibullNP(self.concentration, -1.0 / self.chi)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        k = self.concentration
        one_minus_one_over_k = 1.0 - 1.0 / k
        return one_minus_one_over_k * xp.log(self.chi) - np.euler_gamma * one_minus_one_over_k

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        else:
            shape = self.shape
        lambda_ = self.chi ** (1.0 / self.concentration)
        return jax.random.weibull_min(key, lambda_, self.concentration, shape)



"""These tests ensure that our distributions match scipy's."""
from __future__ import annotations

from functools import partial
from typing import Any

import numpy as np
from jax import Array
from numpy.random import Generator
from numpy.testing import assert_allclose
from tjax import JaxComplexArray, assert_tree_allclose

from efax import (HasEntropyEP, HasEntropyNP, JointDistributionN, MaximumLikelihoodEstimator,
                  Multidimensional, NaturalParametrization, SimpleDistribution,
                  flat_dict_of_observations, flat_dict_of_parameters, parameter_map,
                  unflatten_mapping)

from .create_info import (BetaInfo, ComplexCircularlySymmetricNormalInfo, ComplexNormalInfo,
                          DirichletInfo, IsotropicNormalInfo, MultivariateDiagonalNormalInfo,
                          MultivariateFixedVarianceNormalInfo, MultivariateNormalInfo,
                          VonMisesFisherInfo)
from .distribution_info import DistributionInfo


def test_nat_entropy(generator: Generator,
                     entropy_distribution_info: DistributionInfo[Any, Any, Any]
                     ) -> None:
    """Test that the entropy calculation matches scipy's."""
    shape = (7, 13)
    nat_parameters = entropy_distribution_info.nat_parameter_generator(generator, shape=shape)
    assert isinstance(nat_parameters, HasEntropyNP)
    scipy_distribution = entropy_distribution_info.nat_to_scipy_distribution(nat_parameters)
    rtol = 2e-5
    my_entropy = nat_parameters.entropy()
    scipy_entropy = scipy_distribution.entropy()
    assert_allclose(my_entropy, scipy_entropy, rtol=rtol)


def test_exp_entropy(generator: Generator,
                     entropy_distribution_info: DistributionInfo[Any, Any, Any]
                     ) -> None:
    """Test that the entropy calculation matches scipy's."""
    shape = (7, 13)
    exp_parameters = entropy_distribution_info.exp_parameter_generator(generator, shape=shape)
    assert isinstance(exp_parameters, HasEntropyEP)
    scipy_distribution = entropy_distribution_info.exp_to_scipy_distribution(exp_parameters)
    rtol = (8e-2
            if isinstance(entropy_distribution_info, BetaInfo)
            else 4e-2
            if isinstance(entropy_distribution_info, DirichletInfo)
            else 2e-5)
    my_entropy = exp_parameters.entropy()
    scipy_entropy = scipy_distribution.entropy()
    assert_allclose(my_entropy, scipy_entropy, rtol=rtol)


def check_observation_shape(nat_parameters: NaturalParametrization[Any, Any],
                            efax_x: JaxComplexArray | dict[str, Any]
                            ) -> None:
    """Verify that the sufficient statistics have the right shape."""
    if isinstance(nat_parameters, JointDistributionN):
        assert isinstance(efax_x, dict)
        for name, value in nat_parameters.sub_distributions().items():
            check_observation_shape(value, efax_x[name])
        return
    assert isinstance(nat_parameters, SimpleDistribution)
    assert isinstance(efax_x, Array)
    dimensions = (nat_parameters.dimensions()  # type: ignore[attr-defined]
                  if isinstance(nat_parameters, Multidimensional)  # type: ignore[unreachable]
                  else 0)
    ideal_shape = nat_parameters.domain_support().shape(dimensions)
    assert efax_x.shape == ideal_shape


def test_pdf(generator: Generator, distribution_info: DistributionInfo[Any, Any, Any]) -> None:
    """Test that the density/mass function calculation matches scipy's."""
    for _ in range(10):
        nat_parameters = distribution_info.nat_parameter_generator(generator, shape=())
        scipy_distribution = distribution_info.nat_to_scipy_distribution(nat_parameters)
        scipy_x = scipy_distribution.rvs(random_state=generator)
        efax_x = distribution_info.scipy_to_exp_family_observation(scipy_x)
        check_observation_shape(nat_parameters, efax_x)

        # Verify that the density matches scipy.
        efax_density = np.asarray(nat_parameters.pdf(efax_x), dtype=np.float64)
        try:
            scipy_density = scipy_distribution.pdf(scipy_x)
        except AttributeError:
            scipy_density = scipy_distribution.pmf(scipy_x)

        if isinstance(distribution_info, MultivariateDiagonalNormalInfo):
            atol = 1e-5
            rtol = 3e-4
        else:
            atol = 1e-5
            rtol = 1e-4
        assert_allclose(efax_density, scipy_density, rtol=rtol, atol=atol)


def test_maximum_likelihood_estimation(
        generator: Generator,
        distribution_info: DistributionInfo[NaturalParametrization[Any, Any], Any, Any]
        ) -> None:
    """Test maximum likelihood estimation using SciPy.

    Test that maximum likelihood estimation from scipy-generated variates produce the same
    distribution from which they were drawn.
    """
    if isinstance(distribution_info,
                  ComplexCircularlySymmetricNormalInfo | MultivariateNormalInfo
                  | VonMisesFisherInfo | MultivariateFixedVarianceNormalInfo):
        atol = 2e-2
        rtol = 2e-2
    elif isinstance(distribution_info, ComplexNormalInfo | IsotropicNormalInfo):
        atol = 1e-3
        rtol = 2e-2
    else:
        atol = 1e-4
        rtol = 2e-2
    n = 70000
    # Generate a distribution with expectation parameters.
    exp_parameters = distribution_info.exp_parameter_generator(generator, shape=())
    # Generate variates from the corresponding scipy distribution.
    scipy_distribution = distribution_info.exp_to_scipy_distribution(exp_parameters)
    scipy_x = scipy_distribution.rvs(random_state=generator, size=n)
    # Convert the variates to sufficient statistics.
    efax_x = distribution_info.scipy_to_exp_family_observation(scipy_x)
    flat_efax_x = flat_dict_of_observations(efax_x)
    flat_parameters = flat_dict_of_parameters(exp_parameters)
    flat_efax_x_clamped = {path: flat_parameters[path].domain_support().clamp(value)
                           for path, value in flat_efax_x.items()}
    efax_x_clamped: Array | dict[str, Any] = (flat_efax_x_clamped[()]
                                              if flat_efax_x_clamped.keys() == {()}
                                              else unflatten_mapping(flat_efax_x_clamped))
    estimator = MaximumLikelihoodEstimator.create_estimator(exp_parameters)
    sufficient_stats = estimator.sufficient_statistics(efax_x_clamped)

    # Verify that the mean of the sufficient statistics equals the expectation parameters.
    calculated_parameters = parameter_map(partial(np.mean, axis=0),  # type: ignore[arg-type]
                                          sufficient_stats)
    assert_tree_allclose(exp_parameters, calculated_parameters, rtol=rtol, atol=atol)



from __future__ import annotations

from collections.abc import Generator
from typing import Any

import numpy as np
import pytest
from jax import enable_custom_prng
from jax.experimental import enable_x64
from jax.random import key as jax_key
from numpy.random import Generator as NumpyGenerator
from tjax import KeyArray

from efax import (BooleanRing, HasConjugatePrior, HasEntropyEP, HasEntropyNP,
                  HasGeneralizedConjugatePrior, IntegralRing, JointDistribution, Samplable,
                  Structure)

from .create_info import (BetaInfo, ChiSquareInfo, ComplexCircularlySymmetricNormalInfo,
                          DirichletInfo, GammaInfo, GeneralizedDirichletInfo, JointInfo,
                          create_infos)


@pytest.fixture(autouse=True)
def _jax_enable64() -> Generator[None]:  # pyright: ignore
    with enable_x64():
        yield


def pytest_addoption(parser: Any) -> None:
    parser.addoption('--distribution', action='store', default=None)


@pytest.fixture
def generator() -> NumpyGenerator:
    return np.random.default_rng(123)


@pytest.fixture
def key() -> KeyArray:
    with enable_custom_prng():
        return jax_key(123)


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line(
        "markers", "nondistribution: mark a test as not related to a particular distribution")


def pytest_collection_modifyitems(config: pytest.Config, items: list[pytest.Item]) -> None:
    if config.getoption("--distribution") is None:
        return
    skip_non_distribution = pytest.mark.skip(reason="Distribution selected")
    for item in items:
        if "nondistribution" in item.keywords:
            item.add_marker(skip_non_distribution)


_all_infos = create_infos()


@pytest.fixture
def distribution_name(request: Any) -> str | None:
    return request.config.getoption("--distribution")


def supports(s: Structure[Any], abc: type[Any]) -> bool:
    return all(issubclass(info.type_, abc) or issubclass(info.type_, JointDistribution)
               for info in s.infos)


def any_integral_supports(structure: Structure[Any]) -> bool:
    return any(isinstance(s.ring, BooleanRing | IntegralRing)
               for s in structure.domain_support().values())


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    distribution_name_option = metafunc.config.getoption('--distribution')
    assert isinstance(distribution_name_option, str | type(None))
    if 'distribution_info' in metafunc.fixturenames:
        q = [info
             for info in _all_infos
             if info.tests_selected(distribution_name_option)]
        ids = [info.name() for info in q]
        metafunc.parametrize("distribution_info", q, ids=ids)
    if 'sampling_distribution_info' in metafunc.fixturenames and 'natural' in metafunc.fixturenames:
        p = [(info, natural)
             for info in _all_infos
             for natural in (False, True)
             if supports(info.nat_structure() if natural else info.exp_structure(), Samplable)
             if info.tests_selected(distribution_name_option)]
        ids = [f"{info.name()}{'NP' if natural else 'EP'}" for info, natural in p]
        metafunc.parametrize(("sampling_distribution_info", "natural"), p,
                             ids=ids)
    if ('sampling_wc_distribution_info' in metafunc.fixturenames
            and 'natural' in metafunc.fixturenames):
        p = [(info, natural)
             for info in _all_infos
             for natural in (False, True)
             if supports(structure := (info.nat_structure() if natural else info.exp_structure()),
                         Samplable)
             if not any_integral_supports(structure)
             if not isinstance(info,
                               ComplexCircularlySymmetricNormalInfo | BetaInfo | DirichletInfo
                               | ChiSquareInfo | GammaInfo | JointInfo)
             if info.tests_selected(distribution_name_option)]
        ids = [f"{info.name()}{'NP' if natural else 'EP'}" for info, natural in p]
        metafunc.parametrize(("sampling_wc_distribution_info", "natural"), p,
                             ids=ids)
    if 'cp_distribution_info' in metafunc.fixturenames:
        q = [info
             for info in _all_infos
             if supports(info.exp_structure(), HasConjugatePrior)
             if info.tests_selected(distribution_name_option)]
        metafunc.parametrize("cp_distribution_info", q, ids=[info.name() for info in q])
    if 'gcp_distribution_info' in metafunc.fixturenames:
        q = [info
             for info in _all_infos
             if supports(info.exp_structure(), HasGeneralizedConjugatePrior)
             if info.tests_selected(distribution_name_option)]
        metafunc.parametrize("gcp_distribution_info", q, ids=[info.name() for info in q])
    if 'entropy_distribution_info' in metafunc.fixturenames:
        q = [info
             for info in _all_infos
             if ((supports(info.exp_structure(), HasEntropyEP)
                  or supports(info.nat_structure(), HasEntropyNP))
                 and not isinstance(info, GeneralizedDirichletInfo))
             if info.tests_selected(distribution_name_option)]
        metafunc.parametrize("entropy_distribution_info", q, ids=[info.name() for info in q])



from __future__ import annotations

from typing import Any

import jax.numpy as jnp
import numpy as np
import pytest
from numpy.random import Generator
from tjax import assert_tree_allclose

from efax import Flattener, MultivariateUnitNormalNP

from .distribution_info import DistributionInfo


@pytest.mark.parametrize('natural', [False, True])
def test_flatten(generator: Generator,
                 distribution_info: DistributionInfo[Any, Any, Any],
                 *,
                 natural: bool,
                 ) -> None:
    """Test that unflattening restores the source of flattening."""
    shape = (3, 4)
    p = (distribution_info.exp_parameter_generator(generator, shape=shape)
         if natural
         else distribution_info.nat_parameter_generator(generator, shape=shape))
    flattener, flattened = Flattener.flatten(p)
    assert issubclass(flattened.dtype.type, np.floating)
    assert_tree_allclose(flattener.unflatten(flattened), p)


def test_raises() -> None:
    """Test that unflattening raises for bad shapes."""
    m = MultivariateUnitNormalNP(jnp.zeros(10))
    flattener, flattened = Flattener.flatten(m)
    assert flattened.shape == (10,)
    with pytest.raises(ValueError, match="Incompatible array"):
        flattener.unflatten(jnp.zeros(5))
    with pytest.raises(ValueError, match="Incompatible array"):
        flattener.unflatten(jnp.zeros(12))



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar

from tjax import JaxComplexArray, JaxRealArray, Shape
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..natural_parametrization import EP, NaturalParametrization

TEP = TypeVar('TEP', bound=ExpectationParametrization[Any])
NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])
Domain = TypeVar('Domain', bound=JaxComplexArray)


class TransformedNaturalParametrization(NaturalParametrization[TEP, Domain],
                                        Generic[NP, EP, TEP, Domain]):
    """Produce a NaturalParametrization by relating it to some base distrubtion NP."""
    @classmethod
    @abstractmethod
    def base_distribution_cls(cls) -> type[NP]:
        raise NotImplementedError

    @abstractmethod
    def base_distribution(self) -> NP:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def create_expectation(cls, expectation_parametrization: EP) -> TEP:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def sample_to_base_sample(cls, x: Domain, **fixed_parameters: Any) -> Domain:
        raise NotImplementedError

    @property
    @override
    def shape(self) -> Shape:
        return self.base_distribution().shape

    @override
    def log_normalizer(self) -> JaxRealArray:
        """The log-normalizer."""
        return self.base_distribution().log_normalizer()

    @override
    def to_exp(self) -> TEP:
        """The corresponding expectation parameters."""
        return self.create_expectation(self.base_distribution().to_exp())

    @override
    @classmethod
    def sufficient_statistics(cls, x: Domain, **fixed_parameters: Any) -> TEP:
        y = cls.sample_to_base_sample(x, **fixed_parameters)
        base_cls = cls.base_distribution_cls()
        return cls.create_expectation(base_cls.sufficient_statistics(y, **fixed_parameters))


TNP = TypeVar('TNP', bound=TransformedNaturalParametrization[Any, Any, Any, Any])


class TransformedExpectationParametrization(ExpectationParametrization[TNP],
                                            Generic[EP, NP, TNP]):
    """Produce an ExpectationParametrization by relating it to some base distrubtion EP."""
    @classmethod
    @abstractmethod
    def base_distribution_cls(cls) -> type[EP]:
        raise NotImplementedError

    @abstractmethod
    def base_distribution(self) -> EP:
        raise NotImplementedError

    @abstractmethod
    def create_natural(self, natural_parametrization: NP) -> TNP:
        raise NotImplementedError

    @property
    @override
    def shape(self) -> Shape:
        return self.base_distribution().shape

    @override
    def to_nat(self) -> TNP:
        """The corresponding natural parameters."""
        return self.create_natural(self.base_distribution().to_nat())



from __future__ import annotations

from typing import Any, Generic, TypeVar

import jax
from jax.scipy import special as jss
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..interfaces.multidimensional import Multidimensional
from ..interfaces.samplable import Samplable
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, VectorSupport, distribution_parameter, negative_support

EP = TypeVar('EP', bound='DirichletCommonEP[Any]')


@dataclass
class DirichletCommonNP(HasEntropyNP[EP],
                        Samplable,
                        Multidimensional,
                        NaturalParametrization[EP, JaxRealArray],
                        Generic[EP]):
    alpha_minus_one: JaxRealArray = distribution_parameter(VectorSupport(
        ring=RealField(minimum=-1.0, generation_scale=3.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.alpha_minus_one.shape[:-1]

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        q = self.alpha_minus_one
        return (xp.sum(jss.gammaln(q + 1.0), axis=-1)
                - jss.gammaln(xp.sum(q, axis=-1) + self.dimensions()))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.shape
        return jax.random.dirichlet(key, 1.0 + self.alpha_minus_one, shape)[..., :-1]

    @override
    def dimensions(self) -> int:
        return self.alpha_minus_one.shape[-1]

    def _exp_helper(self) -> JaxRealArray:
        xp = self.get_namespace()
        q = self.alpha_minus_one
        return jss.digamma(q + 1.0) - jss.digamma(xp.sum(q, axis=-1, keepdims=True) + q.shape[-1])


NP = TypeVar('NP', bound=DirichletCommonNP[Any])


@dataclass
class DirichletCommonEP(HasEntropyEP[NP],
                        Samplable,
                        ExpToNat[NP],
                        Multidimensional, Generic[NP]):
    mean_log_probability: JaxRealArray = distribution_parameter(VectorSupport(
        ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_log_probability.shape[:-1]

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def dimensions(self) -> int:
        return self.mean_log_probability.shape[-1]

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_nat().sample(key, shape)



from __future__ import annotations

from abc import abstractmethod
from typing import Any, Generic, TypeVar, final

from tjax import JaxRealArray, jit

from .natural_parametrization import NaturalParametrization
from .parametrization import Distribution
from .tools import parameter_dot_product, parameter_map

NP = TypeVar('NP', bound=NaturalParametrization[Any, Any])


class ExpectationParametrization(Distribution, Generic[NP]):
    """The expectation parametrization of an exponential family distribution.

    This class also doubles as the sufficient statistics of an exponential family distribution.

    The motivation for the expectation parametrization is combining independent and identically
    distributed observations into the maximum likelihood distribution.  In the expectation
    parametrization, this is an expected value.
    """
    @classmethod
    @abstractmethod
    def natural_parametrization_cls(cls) -> type[NP]:
        raise NotImplementedError

    @abstractmethod
    def to_nat(self) -> NP:
        """The corresponding natural parameters."""
        raise NotImplementedError

    @jit
    @final
    def kl_divergence(self, q: NP, *, self_nat: NP | None = None) -> JaxRealArray:
        """The Kullback-Leibler divergence.

        This can be quite slow since it depends on a conversion to natural parameters.

        Args:
            self: The expectation parameters of the first distribution.
            q: The natural parameters of second destribution.
            self_nat: The natural parameters of the first distribution, if available.
        """
        if self_nat is None:
            self_nat = self.to_nat()
        xp = self.get_namespace()
        difference = parameter_map(xp.subtract, self_nat, q)
        return (parameter_dot_product(difference, self)
                + q.log_normalizer()
                - self_nat.log_normalizer())



from __future__ import annotations

import math
from typing import Any, Self

import jax
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.conjugate_prior import HasGeneralizedConjugatePrior
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import RealField, ScalarSupport, VectorSupport, distribution_parameter
from .diagonal import MultivariateDiagonalNormalNP
from .isotropic import IsotropicNormalNP


@dataclass
class MultivariateFixedVarianceNormalNP(HasEntropyNP['MultivariateFixedVarianceNormalEP'],
                                        NaturalParametrization['MultivariateFixedVarianceNormalEP',
                                                               JaxRealArray],
                                        Multidimensional,
                                        Samplable):
    """The natural parametrization of the multivariate normal distribution with fixed variance.

    This is a curved exponential family.

    Args:
        mean_times_precision: E(x) / Var(x).
        variance: The fixed variance, Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(ScalarSupport(ring=RealField(minimum=0.0)),
                                                    fixed=True)

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        eta = self.mean_times_precision
        return 0.5 * (xp.sum(xp.square(eta), axis=-1) * self.variance
                      + self.dimensions() * xp.log(math.pi * 2.0 * self.variance))

    @override
    def to_exp(self) -> MultivariateFixedVarianceNormalEP:
        xp = self.get_namespace()
        return MultivariateFixedVarianceNormalEP(
            self.mean_times_precision * self.variance[..., xp.newaxis],
            variance=self.variance)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return -0.5 * xp.sum(xp.square(x), axis=-1) / self.variance

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateFixedVarianceNormalEP:
        return MultivariateFixedVarianceNormalEP(x, variance=fixed_parameters['variance'])

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]


@dataclass
class MultivariateFixedVarianceNormalEP(
        HasEntropyEP[MultivariateFixedVarianceNormalNP],
        HasGeneralizedConjugatePrior,
        Samplable,
        ExpectationParametrization[MultivariateFixedVarianceNormalNP],
        Multidimensional):
    """The expectation parametrization of the multivariate normal distribution with fixed variance.

    This is a curved exponential family.

    Args:
        mean: E(x).
        variance: The fixed variance, Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(ScalarSupport(ring=RealField(minimum=0.0)),
                                                    fixed=True)

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateFixedVarianceNormalNP]:
        return MultivariateFixedVarianceNormalNP

    @override
    def to_nat(self) -> MultivariateFixedVarianceNormalNP:
        xp = self.get_namespace()
        return MultivariateFixedVarianceNormalNP(self.mean / self.variance[..., xp.newaxis],
                                                 variance=self.variance)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return -0.5 * (xp.sum(xp.square(self.mean), axis=-1) / self.variance + self.dimensions())

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        xp = self.get_namespace()
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        variance = self.variance[..., xp.newaxis]
        deviation = xp.sqrt(variance)
        return jax.random.normal(key, shape) * deviation + self.mean

    @override
    def conjugate_prior_distribution(self, n: JaxRealArray) -> IsotropicNormalNP:
        xp = self.get_namespace()
        n_over_variance = n / self.variance
        negative_half_precision = -0.5 * n_over_variance
        return IsotropicNormalNP(n_over_variance[..., xp.newaxis] * self.mean,
                                 negative_half_precision)

    @classmethod
    @override
    def from_conjugate_prior_distribution(cls, cp: NaturalParametrization[Any, Any],
                                          variance: JaxRealArray | None = None
                                          ) -> tuple[Self, JaxRealArray]:
        assert isinstance(cp, IsotropicNormalNP)
        assert variance is not None
        xp = cp.get_namespace()
        n_over_variance = -2.0 * cp.negative_half_precision
        n = n_over_variance * variance
        mean = cp.mean_times_precision / n_over_variance[..., xp.newaxis]
        return cls(mean, variance), n

    @override
    def generalized_conjugate_prior_distribution(self, n: JaxRealArray
                                                 ) -> MultivariateDiagonalNormalNP:
        xp = self.get_namespace()
        n_over_variance = n / self.variance[..., xp.newaxis]
        negative_half_precision = -0.5 * n_over_variance
        return MultivariateDiagonalNormalNP(n_over_variance * self.mean, negative_half_precision)

    @override
    def conjugate_prior_observation(self) -> JaxRealArray:
        return self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]



from __future__ import annotations

from collections.abc import Callable
from typing import Any, TypeAlias

import pytest
from jax import grad, jvp, vjp
from jax.custom_derivatives import zero_from_primal
from numpy.random import Generator
from numpy.testing import assert_allclose
from tjax import JaxRealArray, assert_tree_allclose, jit, print_generic, tree_allclose

from efax import NaturalParametrization, Structure, parameters

from .create_info import GeneralizedDirichletInfo
from .distribution_info import DistributionInfo

LogNormalizer: TypeAlias = Callable[[NaturalParametrization[Any, Any]], JaxRealArray]


def test_conversion(generator: Generator,
                    distribution_info: DistributionInfo[Any, Any, Any]
                    ) -> None:
    """Test that the conversion between the different parametrizations are consistent."""
    if isinstance(distribution_info, GeneralizedDirichletInfo):
        pytest.skip()

    n = 30
    shape = (n,)
    original_np = distribution_info.nat_parameter_generator(generator, shape=shape)
    intermediate_ep = original_np.to_exp()
    final_np = intermediate_ep.to_nat()

    # Check round trip.
    if not tree_allclose(final_np, original_np):
        for i in range(n):
            if not tree_allclose(final_np[i], original_np[i]):
                print_generic(original=original_np[i], intermediate=intermediate_ep[i])
        pytest.fail("Conversion failure")

    # Check fixed parameters.
    original_fixed = parameters(original_np, fixed=True)
    intermediate_fixed = parameters(intermediate_ep, fixed=True)
    final_fixed = parameters(final_np, fixed=True)
    assert_tree_allclose(original_fixed, intermediate_fixed)
    assert_tree_allclose(original_fixed, final_fixed)


def prelude(generator: Generator,
            distribution_info: DistributionInfo[Any, Any, Any]
            ) -> tuple[LogNormalizer, LogNormalizer]:
    cls = type(distribution_info.nat_parameter_generator(generator, shape=()))
    original_ln = cls._original_log_normalizer
    optimized_ln = cls.log_normalizer
    return original_ln, optimized_ln


def test_gradient_log_normalizer_primals(generator: Generator,
                                         distribution_info: DistributionInfo[Any, Any, Any]
                                         ) -> None:
    """Tests that the gradient log-normalizer equals the gradient of the log-normalizer."""
    original_ln, optimized_ln = prelude(generator, distribution_info)
    original_gln = jit(grad(original_ln, allow_int=True))
    optimized_gln = jit(grad(optimized_ln, allow_int=True))
    for _ in range(20):
        generated_np = distribution_info.nat_parameter_generator(generator, shape=())
        generated_ep = generated_np.to_exp()  # Regular transformation.
        generated_parameters = parameters(generated_ep, fixed=False)
        structure_ep = Structure.create(generated_ep)

        # Original GLN.
        origianl_gln_np = original_gln(generated_np)
        original_gln_ep = structure_ep.reinterpret(origianl_gln_np)
        original_gln_parameters = parameters(original_gln_ep, fixed=False)

        # Optimized GLN.
        optimized_gln_np = optimized_gln(generated_np)
        optimized_gln_ep = structure_ep.reinterpret(optimized_gln_np)
        optimized_gln_parameters = parameters(optimized_gln_ep, fixed=False)

        # Test primal evaluation.
        # parameters(generated_ep, fixed=False)
        assert_tree_allclose(generated_parameters, original_gln_parameters, rtol=1e-5)
        assert_tree_allclose(generated_parameters, optimized_gln_parameters, rtol=1e-5)


def unit_tangent(nat_parameters: NaturalParametrization[Any, Any]
                 ) -> NaturalParametrization[Any, Any]:
    xp = nat_parameters.get_namespace()
    new_variable_parameters = {path: xp.ones_like(value)
                               for path, value in parameters(nat_parameters, fixed=False).items()}
    new_fixed_parameters = {path: zero_from_primal(value, symbolic_zeros=False)
                            for path, value in parameters(nat_parameters, fixed=True).items()}
    structure = Structure.create(nat_parameters)
    return structure.assemble({**new_variable_parameters, **new_fixed_parameters})


def test_gradient_log_normalizer_jvp(generator: Generator,
                                     distribution_info: DistributionInfo[Any, Any, Any]
                                     ) -> None:
    """Tests that the gradient log-normalizer equals the gradient of the log-normalizer."""
    original_ln, optimized_ln = prelude(generator, distribution_info)
    for _ in range(20):
        nat_parameters = distribution_info.nat_parameter_generator(generator, shape=())

        # Test JVP.
        nat_tangent = unit_tangent(nat_parameters)
        original_ln_of_nat, original_jvp = jvp(original_ln, (nat_parameters,), (nat_tangent,))
        optimized_ln_of_nat, optimized_jvp = jvp(optimized_ln, (nat_parameters,), (nat_tangent,))
        assert_allclose(original_ln_of_nat, optimized_ln_of_nat, rtol=1.5e-5)
        assert_allclose(original_jvp, optimized_jvp, rtol=1.5e-5)


def test_gradient_log_normalizer_vjp(generator: Generator,
                                     distribution_info: DistributionInfo[Any, Any, Any]
                                     ) -> None:
    """Tests that the gradient log-normalizer equals the gradient of the log-normalizer."""
    original_ln, optimized_ln = prelude(generator, distribution_info)
    for _ in range(20):
        nat_parameters = distribution_info.nat_parameter_generator(generator, shape=())
        nat_tangent = unit_tangent(nat_parameters)
        original_ln_of_nat, _ = jvp(original_ln, (nat_parameters,), (nat_tangent,))
        original_ln_of_nat_b, original_vjp = vjp(original_ln, nat_parameters)
        original_gln_of_nat, = original_vjp(1.0)
        optimized_ln_of_nat_b, optimized_vjp = vjp(optimized_ln, nat_parameters)
        optimized_gln_of_nat, = optimized_vjp(1.0)

        assert_allclose(original_ln_of_nat_b, optimized_ln_of_nat_b, rtol=1e-5)
        assert_allclose(original_ln_of_nat, original_ln_of_nat_b, rtol=1e-5)
        assert_tree_allclose(parameters(original_gln_of_nat, fixed=False),
                             parameters(optimized_gln_of_nat, fixed=False),
                             rtol=1e-5)



"""The EFAX Library."""
from ._src.distributions.bernoulli import BernoulliEP, BernoulliNP
from ._src.distributions.beta import BetaEP, BetaNP
from ._src.distributions.chi import ChiEP, ChiNP
from ._src.distributions.chi_square import ChiSquareEP, ChiSquareNP
from ._src.distributions.cmvn.circularly_symmetric import (ComplexCircularlySymmetricNormalEP,
                                                           ComplexCircularlySymmetricNormalNP)
from ._src.distributions.cmvn.unit import (ComplexMultivariateUnitNormalEP,
                                           ComplexMultivariateUnitNormalNP)
from ._src.distributions.complex_normal.complex_normal import ComplexNormalEP, ComplexNormalNP
from ._src.distributions.complex_normal.unit import ComplexUnitNormalEP, ComplexUnitNormalNP
from ._src.distributions.dirichlet import DirichletEP, DirichletNP
from ._src.distributions.exponential import ExponentialEP, ExponentialNP
from ._src.distributions.gamma import GammaEP, GammaNP, GammaVP
from ._src.distributions.gen_dirichlet import GeneralizedDirichletEP, GeneralizedDirichletNP
from ._src.distributions.geometric import GeometricEP, GeometricNP
from ._src.distributions.logarithmic import LogarithmicEP, LogarithmicNP
from ._src.distributions.multinomial import MultinomialEP, MultinomialNP
from ._src.distributions.multivariate_normal.arbitrary import (MultivariateNormalEP,
                                                               MultivariateNormalNP,
                                                               MultivariateNormalVP)
from ._src.distributions.multivariate_normal.diagonal import (MultivariateDiagonalNormalEP,
                                                              MultivariateDiagonalNormalNP,
                                                              MultivariateDiagonalNormalVP)
from ._src.distributions.multivariate_normal.fixed_variance import (
    MultivariateFixedVarianceNormalEP, MultivariateFixedVarianceNormalNP)
from ._src.distributions.multivariate_normal.isotropic import IsotropicNormalEP, IsotropicNormalNP
from ._src.distributions.multivariate_normal.unit import (MultivariateUnitNormalEP,
                                                          MultivariateUnitNormalNP)
from ._src.distributions.negative_binomial import NegativeBinomialEP, NegativeBinomialNP
from ._src.distributions.normal.normal import NormalEP, NormalNP, NormalVP
from ._src.distributions.normal.unit import UnitNormalEP, UnitNormalNP
from ._src.distributions.poisson import PoissonEP, PoissonNP
from ._src.distributions.rayleigh import RayleighEP, RayleighNP
from ._src.distributions.von_mises import VonMisesFisherEP, VonMisesFisherNP
from ._src.distributions.weibull import WeibullEP, WeibullNP
from ._src.expectation_parametrization import ExpectationParametrization
from ._src.interfaces.conjugate_prior import HasConjugatePrior, HasGeneralizedConjugatePrior
from ._src.interfaces.multidimensional import Multidimensional
from ._src.interfaces.samplable import Samplable
from ._src.iteration import (flat_dict_of_observations, flat_dict_of_parameters, flatten_mapping,
                             parameters, support, unflatten_mapping)
from ._src.mixins.has_entropy import HasEntropy, HasEntropyEP, HasEntropyNP
from ._src.natural_parametrization import NaturalParametrization
from ._src.parameter import (BooleanRing, ComplexField, IntegralRing, RealField, Ring,
                             ScalarSupport, SquareMatrixSupport, Support, SymmetricMatrixSupport,
                             VectorSupport)
from ._src.parametrization import Distribution, SimpleDistribution
from ._src.scipy_replacement.complex_multivariate_normal import ScipyComplexMultivariateNormal
from ._src.scipy_replacement.complex_normal import ScipyComplexNormal
from ._src.scipy_replacement.dirichlet import ScipyDirichlet, ScipyGeneralizedDirichlet
from ._src.scipy_replacement.geometric import ScipyGeometric
from ._src.scipy_replacement.joint import ScipyJointDistribution
from ._src.scipy_replacement.multivariate_normal import ScipyMultivariateNormal
from ._src.scipy_replacement.von_mises import ScipyVonMises, ScipyVonMisesFisher
from ._src.structure import Flattener, MaximumLikelihoodEstimator, Structure, SubDistributionInfo
from ._src.tools import parameter_dot_product, parameter_map, parameter_mean
from ._src.transform.joint import JointDistribution, JointDistributionE, JointDistributionN

__all__ = [
    'BernoulliEP',
    'BernoulliNP',
    'BetaEP',
    'BetaNP',
    'BooleanRing',
    'ChiEP',
    'ChiNP',
    'ChiSquareEP',
    'ChiSquareNP',
    'ComplexCircularlySymmetricNormalEP',
    'ComplexCircularlySymmetricNormalNP',
    'ComplexField',
    'ComplexMultivariateUnitNormalEP',
    'ComplexMultivariateUnitNormalNP',
    'ComplexNormalEP',
    'ComplexNormalNP',
    'ComplexUnitNormalEP',
    'ComplexUnitNormalNP',
    'DirichletEP',
    'DirichletNP',
    'Distribution',
    'ExpectationParametrization',
    'ExponentialEP',
    'ExponentialNP',
    'Flattener',
    'GammaEP',
    'GammaNP',
    'GammaVP',
    'GeneralizedDirichletEP',
    'GeneralizedDirichletNP',
    'GeometricEP',
    'GeometricNP',
    'HasConjugatePrior',
    'HasEntropy',
    'HasEntropyEP',
    'HasEntropyNP',
    'HasGeneralizedConjugatePrior',
    'IntegralRing',
    'IsotropicNormalEP',
    'IsotropicNormalNP',
    'JointDistribution',
    'JointDistributionE',
    'JointDistributionN',
    'LogarithmicEP',
    'LogarithmicNP',
    'MaximumLikelihoodEstimator',
    'Multidimensional',
    'MultinomialEP',
    'MultinomialNP',
    'MultivariateDiagonalNormalEP',
    'MultivariateDiagonalNormalNP',
    'MultivariateDiagonalNormalVP',
    'MultivariateFixedVarianceNormalEP',
    'MultivariateFixedVarianceNormalNP',
    'MultivariateNormalEP',
    'MultivariateNormalNP',
    'MultivariateNormalVP',
    'MultivariateUnitNormalEP',
    'MultivariateUnitNormalNP',
    'NaturalParametrization',
    'NegativeBinomialEP',
    'NegativeBinomialNP',
    'NormalEP',
    'NormalNP',
    'NormalVP',
    'PoissonEP',
    'PoissonNP',
    'RayleighEP',
    'RayleighNP',
    'RealField',
    'Ring',
    'Samplable',
    'ScalarSupport',
    'ScipyComplexMultivariateNormal',
    'ScipyComplexNormal',
    'ScipyDirichlet',
    'ScipyGeneralizedDirichlet',
    'ScipyGeometric',
    'ScipyJointDistribution',
    'ScipyMultivariateNormal',
    'ScipyVonMises',
    'ScipyVonMisesFisher',
    'SimpleDistribution',
    'SquareMatrixSupport',
    'Structure',
    'SubDistributionInfo',
    'Support',
    'SymmetricMatrixSupport',
    'UnitNormalEP',
    'UnitNormalNP',
    'VectorSupport',
    'VonMisesFisherEP',
    'VonMisesFisherNP',
    'WeibullEP',
    'WeibullNP',
    'flat_dict_of_observations',
    'flat_dict_of_parameters',
    'flatten_mapping',
    'parameter_dot_product',
    'parameter_map',
    'parameter_mean',
    'parameters',
    'support',
    'unflatten_mapping',
]



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Self, TypeVar, final, get_type_hints

from jax import grad, jacfwd, vjp, vmap
from tjax import (JaxAbstractClass, JaxComplexArray, JaxRealArray, abstract_custom_jvp,
                  abstract_jit, jit)
from tjax.dataclasses import dataclass

from .iteration import parameters
from .parametrization import Distribution
from .structure import Flattener, MaximumLikelihoodEstimator, Structure
from .tools import parameter_dot_product

if TYPE_CHECKING:
    from .expectation_parametrization import ExpectationParametrization


EP = TypeVar('EP', bound='ExpectationParametrization[Any]')
Domain = TypeVar('Domain', bound=JaxComplexArray | dict[str, Any])


def log_normalizer_jvp(primals: tuple[NaturalParametrization[Any, Any]],
                       tangents: tuple[NaturalParametrization[Any, Any]],
                       ) -> tuple[JaxRealArray, JaxRealArray]:
    """The log-normalizer's special JVP vastly improves numerical stability."""
    q, = primals
    q_dot, = tangents
    y = q.log_normalizer()
    p = q.to_exp()
    y_dot = parameter_dot_product(q_dot, p)
    return y, y_dot


@dataclass
class NaturalParametrization(Distribution,
                             JaxAbstractClass,
                             Generic[EP, Domain]):
    """The natural parametrization of an exponential family distribution.

    The motivation for the natural parametrization is combining and scaling independent predictive
    evidence.  In the natural parametrization, these operations correspond to scaling and addition.
    """
    @abstract_custom_jvp(log_normalizer_jvp)
    @abstract_jit
    @abstractmethod
    def log_normalizer(self) -> JaxRealArray:
        """Returns: The log-normalizer."""
        raise NotImplementedError

    @abstract_jit
    @abstractmethod
    def to_exp(self) -> EP:
        """Returns: The corresponding expectation parameters."""
        raise NotImplementedError

    @abstract_jit
    @abstractmethod
    def carrier_measure(self, x: Domain) -> JaxRealArray:
        """The corresponding carrier measure.

        Args:
            x: The sample.
        Returns: The corresponding carrier measure, which is typically xp.zeros(x.shape[:-1]).
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def sufficient_statistics(cls, x: Domain, **fixed_parameters: Any) -> EP:
        """The sufficient statistics corresponding to an observation.

        This is typically used in maximum likelihood estimation.

        Args:
            x: The samples.
            fixed_parameters: The fixed parameters of the expectation parametrization.
        Returns: The sufficient statistics, stored as expectation parameters.
        """
        raise NotImplementedError

    @classmethod
    def expectation_parametrization_cls(cls) -> type[EP]:
        return get_type_hints(cls.to_exp)['return']

    @jit
    @final
    def pdf(self, x: Domain) -> JaxRealArray:
        """The distribution's density or mass function at x.

        Args:
            x: The sample.
        """
        xp = self.get_namespace()
        return xp.exp(self.log_pdf(x))

    @jit
    @final
    def log_pdf(self, x: Domain) -> JaxRealArray:
        """The distribution's density or mass function at x.

        Args:
            x: The sample.
        """
        estimator = MaximumLikelihoodEstimator.create_estimator_from_natural(self)
        tx = estimator.sufficient_statistics(x)
        return parameter_dot_product(self, tx) - self.log_normalizer() + self.carrier_measure(x)

    @final
    def fisher_information_diagonal(self) -> Self:
        """The diagonal elements of the Fisher information.

        Returns: The Fisher information stored in a NaturalParametrization object whose fields
            are an array of the same shape as self.

        See also: apply_fisher_information.
        """
        xp = self.get_namespace()
        flattener, _ = Flattener.flatten(self)
        fisher_matrix = self._fisher_information_matrix()
        fisher_diagonal = xp.linalg.diagonal(fisher_matrix)
        return flattener.unflatten(fisher_diagonal)

    @final
    def fisher_information_trace(self) -> Self:
        """The trace of the Fisher information.

        Returns: The trace of the Fisher information stored in a NaturalParametrization object whose
            fields are scalar.

        See also: apply_fisher_information.
        """
        xp = self.get_namespace()
        fisher_information_diagonal = self.fisher_information_diagonal()
        structure = Structure.create(self)
        final_parameters = parameters(self)
        for path, (value, support) in parameters(fisher_information_diagonal, support=True).items():
            na = support.axes()
            if na == 0:
                new_value = value
            elif na == 1:
                new_value = xp.sum(value, axis=-1)
            elif na == 2:  # noqa: PLR2004
                new_value = xp.sum(xp.triu(value), axis=(-2, -1))
            else:
                raise RuntimeError
            final_parameters[path] = new_value
        return structure.assemble(final_parameters)

    @final
    def jeffreys_prior(self) -> JaxRealArray:
        xp = self.get_namespace()
        fisher_matrix = self._fisher_information_matrix()
        return xp.sqrt(xp.linalg.det(fisher_matrix))

    @jit
    @final
    def apply_fisher_information(self, vector: EP) -> tuple[EP, Self]:
        """Efficiently apply the Fisher information matrix to a vector.

        Args:
            vector: Some set of expectation parameters.

        Returns:
            The expectation parameters corresponding to self.
            The Fisher information of self applied to the inputted vector.
        """
        expectation_parameters: EP
        expectation_parameters, f_vjp = vjp(type(self).to_exp, self)
        return expectation_parameters, f_vjp(vector)

    @final
    def kl_divergence(self, q: Self) -> JaxRealArray:
        return self.to_exp().kl_divergence(q, self_nat=self)

    @classmethod
    def _flat_log_normalizer(cls,
                             flattened_parameters: JaxRealArray,
                             flattener: Flattener[Self],
                             ) -> JaxRealArray:
        distribution = flattener.unflatten(flattened_parameters)
        return distribution.log_normalizer()

    @jit
    def _fisher_information_matrix(self) -> JaxRealArray:
        flattener, flattened = Flattener.flatten(self, map_to_plane=False)
        fisher_info_f = jacfwd(grad(self._flat_log_normalizer))
        for _ in range(len(self.shape)):
            fisher_info_f = vmap(fisher_info_f)
        return fisher_info_f(flattened, flattener)



from __future__ import annotations

from typing import Any

import numpy as np
from numpy.random import Generator

from efax import HasEntropyEP, HasEntropyNP, SimpleDistribution, parameters

from .distribution_info import DistributionInfo


def test_shapes(generator: Generator, distribution_info: DistributionInfo[Any, Any, Any]) -> None:
    """Test that the methods produce the correct shapes."""
    shape = (3, 4)

    p = distribution_info.exp_parameter_generator(generator, shape=shape)
    q = distribution_info.nat_parameter_generator(generator, shape=shape)
    scipy_dist = distribution_info.exp_to_scipy_distribution(p)
    scipy_x = scipy_dist.rvs()
    x = distribution_info.scipy_to_exp_family_observation(scipy_x)

    def check(z: SimpleDistribution) -> None:
        for xf, support in parameters(z, support=True).values():
            assert xf.shape[:len(xf.shape) - support.axes()] == shape

    assert p.shape == shape
    assert q.shape == shape

    check(p)
    check(q)

    assert q.log_normalizer().shape == shape
    if isinstance(p, HasEntropyEP):
        assert p.cross_entropy(q).shape == shape
    if isinstance(q, HasEntropyNP):
        assert q.entropy().shape == shape
    assert q.carrier_measure(x).shape == shape
    if isinstance(p, HasEntropyEP):
        assert p.expected_carrier_measure().shape == shape
    assert q.pdf(x).shape == shape


def test_types(distribution_info: DistributionInfo[Any, Any, Any]) -> None:
    if isinstance(distribution_info.exp_parameter_generator(np.random.default_rng(), ()), tuple):
        msg = "This should return a number or an ndarray"
        raise TypeError(msg)



from __future__ import annotations

from typing import Any

from tjax import JaxRealArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..natural_parametrization import NaturalParametrization
from ..parameter import (IntegralRing, RealField, ScalarSupport, distribution_parameter,
                         negative_support)
from ..parametrization import SimpleDistribution

log_probability_floor = -50.0
log_probability_ceiling = -1e-7


@dataclass
class LogarithmicNP(NaturalParametrization['LogarithmicEP', JaxRealArray],
                    SimpleDistribution):
    """The natural parametrization of the logarithmic distribution.

    Args:
        log_probability: log(p).
    """
    log_probability: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.log_probability.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=IntegralRing(minimum=1))

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.log(-xp.log1p(-xp.exp(self.log_probability)))

    @override
    def to_exp(self) -> LogarithmicEP:
        xp = self.get_namespace()
        probability = xp.exp(self.log_probability)
        chi = xp.where(self.log_probability < log_probability_floor,
                       xp.asarray(1.0),
                       xp.where(self.log_probability > log_probability_ceiling,
                                xp.asarray(xp.inf),
                                probability / (xp.expm1(self.log_probability)
                                               * xp.log1p(-probability))))
        return LogarithmicEP(chi)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return -xp.log(x)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> LogarithmicEP:
        return LogarithmicEP(x)


@dataclass
class LogarithmicEP(ExpToNat[LogarithmicNP],
                    ExpectationParametrization[LogarithmicNP],
                    SimpleDistribution):
    """The expectation parametrization of the logarithmic distribution.

    Args:
        chi: -(p / (1-p)) * log(1-p).
    """
    chi: JaxRealArray = distribution_parameter(ScalarSupport(ring=RealField(minimum=1.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.chi.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=IntegralRing(minimum=1))

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[LogarithmicNP]:
        return LogarithmicNP

    # The expected_carrier_measure is unknown.

    @override
    def to_nat(self) -> LogarithmicNP:
        xp = self.get_namespace()
        z: LogarithmicNP = super().to_nat()
        return LogarithmicNP(xp.where(self.chi < 1.0,
                                      xp.asarray(xp.nan),
                                      xp.where(self.chi == 1.0,
                                               xp.asarray(xp.inf),
                                               z.log_probability)))



from __future__ import annotations

import numpy as np
import pytest
from numpy.random import Generator
from numpy.testing import assert_allclose
from tjax import NumpyComplexArray, NumpyRealArray, Shape

from efax import ScipyComplexMultivariateNormal, ScipyComplexNormal


# Tools --------------------------------------------------------------------------------------------
def random_complex_array(generator: Generator, shape: Shape = ()) -> NumpyComplexArray:
    return np.asarray(sum(x * generator.normal(size=shape) for x in (0.5, 0.5j)))


def build_uvcn(generator: Generator, shape: Shape) -> ScipyComplexNormal:
    mean = random_complex_array(generator, shape)
    variance = generator.exponential(size=shape)
    pseudo_variance = (variance
                       * generator.beta(2, 2, size=shape)
                       * np.exp(1j * generator.uniform(0, 2 * np.pi, size=shape)))
    return ScipyComplexNormal(mean, variance, pseudo_variance)


def build_mvcn(generator: Generator,
               shape: Shape,
               dimensions: int,
               polarization: float = 0.98
               ) -> ScipyComplexMultivariateNormal:
    directions = 3
    weights = np.asarray(range(directions)) + 1.5
    mean = random_complex_array(generator, (*shape, dimensions))
    z = random_complex_array(generator, (*shape, dimensions, directions))
    regularizer = np.tile(np.eye(dimensions), (*shape, 1, 1))
    variance = (
        np.average(np.conj(z)[..., np.newaxis, :, :] * z[..., np.newaxis, :],
                   weights=weights,
                   axis=-1)
        + regularizer)
    pseudo_variance = np.average(z[..., np.newaxis, :, :] * z[..., np.newaxis, :],
                                 weights=weights,
                                 axis=-1) * polarization
    return ScipyComplexMultivariateNormal(mean, variance, pseudo_variance)


# Tests --------------------------------------------------------------------------------------------
@pytest.mark.nondistribution
def test_univariate_rvs(generator: Generator) -> None:
    shape = (3, 2)
    rvs_shape = (50, 100)
    rvs_axes = tuple(range(-len(rvs_shape), 0))
    dist = build_uvcn(generator, shape)
    rvs = dist.rvs(random_state=generator, size=rvs_shape)
    assert rvs.shape == shape + rvs_shape

    estimated_mean = np.average(rvs, axis=rvs_axes)
    centered_rvs = rvs - dist.mean[(...,) + (np.newaxis,) * len(rvs_shape)]
    estimated_variance = np.real(np.average(
        np.conj(centered_rvs) * centered_rvs, axis=rvs_axes))
    estimated_pseudo_variance = np.average(np.square(centered_rvs), axis=rvs_axes)
    assert_allclose(estimated_mean, dist.mean, rtol=0.0, atol=2e-2)
    assert_allclose(estimated_variance, dist.variance, rtol=3e-2, atol=2e-2)
    assert_allclose(estimated_pseudo_variance, dist.pseudo_variance, rtol=3e-2,
                    atol=2e-2)


@pytest.mark.nondistribution
def test_multivariate_rvs(generator: Generator) -> None:
    shape = (3, 5)
    rvs_shape = (1800, 1700)
    dimensions = 2
    rvs_axes = tuple(range(-len(rvs_shape) - 1, -1))
    rvs_axes2 = tuple(range(-len(rvs_shape) - 2, -2))
    dist = build_mvcn(generator, shape, dimensions)
    rvs = dist.rvs(random_state=generator, size=rvs_shape)
    assert rvs.shape == shape + rvs_shape + (dimensions,)

    estimated_mean = np.average(rvs, axis=rvs_axes)
    centered_rvs = rvs - dist.mean[(...,) + (np.newaxis,) * len(rvs_shape) + (slice(None),)]
    estimated_variance = np.real(np.average(
        centered_rvs[..., np.newaxis] * np.conj(centered_rvs)[..., np.newaxis, :],
        axis=rvs_axes2))
    estimated_pseudo_variance = np.average(
        centered_rvs[..., np.newaxis] * centered_rvs[..., np.newaxis, :], axis=rvs_axes2)
    assert_allclose(estimated_mean, dist.mean, rtol=1e-2, atol=1e-2)
    assert_allclose(estimated_variance, dist.variance, rtol=2e-1, atol=5e-1)
    assert_allclose(estimated_pseudo_variance, dist.pseudo_variance, rtol=2e-1,
                    atol=5e-1)


@pytest.mark.nondistribution
def test_univariate_multivariate_consistency(generator: Generator) -> None:
    mv = build_mvcn(generator, (), 1, polarization=0.5)
    component = mv.access_object(())
    mean: NumpyComplexArray = np.asarray(component.mean[0])
    variance: NumpyRealArray = np.real(np.asarray(component.variance[0, 0]))
    pseudo_variance: NumpyComplexArray = np.asarray(component.pseudo_variance[0, 0])
    uv = ScipyComplexNormal(mean, variance, pseudo_variance)
    x = random_complex_array(generator)
    assert_allclose(mv.pdf(x[np.newaxis]), uv.pdf(x))



from __future__ import annotations

from typing import Any, Self

import numpy as np
import optype.numpy as onp
import scipy.stats as ss
from tjax import NumpyRealArray
from typing_extensions import override

from .shaped_distribution import ShapedDistribution

mvn: type = ss._multivariate.multivariate_normal_frozen  # noqa: SLF001


class ScipyMultivariateNormalUnvectorized(mvn):
    """This class repairs multivariate_normal.

    See https://github.com/scipy/scipy/issues/7689.
    """
    @override
    def rvs(self,
            size: Any = None,
            random_state: Any = None) -> onp.ArrayND[np.float64]:
        retval = super().rvs(size=size, random_state=random_state)
        if size is None:
            size = ()
        elif isinstance(size, int):
            size = (size,)
        else:
            size = tuple(size)
        return np.reshape(retval, size + self.mean.shape)


class ScipyMultivariateNormal(ShapedDistribution[ScipyMultivariateNormalUnvectorized]):
    """This class allows distributions having a non-empty shape."""
    @classmethod
    def from_mc(cls,
                mean: NumpyRealArray | None = None,
                cov: NumpyRealArray | None = None) -> Self:
        if mean is None and cov is None:
            mean = np.zeros(1)
        if mean is None:
            if cov is None:
                raise ValueError
            mean = np.zeros(cov.shape[:-1], dtype=cov.dtype)
        if cov is None:
            cov = np.tile(np.eye(mean.shape[-1]), mean.shape[:-1] + (1, 1))
        shape = mean[..., -1].shape
        rvs_shape = (mean.shape[-1],)
        dtype = mean.dtype
        objects = np.empty(shape, dtype=ScipyMultivariateNormalUnvectorized)
        for i in np.ndindex(*shape):
            objects[i] = ScipyMultivariateNormalUnvectorized(mean[i], cov[i])
        return cls(shape, rvs_shape, dtype, objects)



from __future__ import annotations

from typing import Any

import jax
import numpy as np
from array_api_compat import get_namespace
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import VectorSupport, distribution_parameter, negative_support, positive_support


@dataclass
class MultivariateDiagonalNormalNP(HasEntropyNP['MultivariateDiagonalNormalEP'],
                                   NaturalParametrization['MultivariateDiagonalNormalEP',
                                                          JaxRealArray],
                                   Multidimensional,
                                   Samplable):
    """The natural parametrization of the normal distribution with diagonal variance.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_half_precision: The diagonal elements of -0.5 / Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(VectorSupport(
        ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_times_precision.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        components = (-xp.square(self.mean_times_precision) / (4.0 * self.negative_half_precision)
                      + 0.5 * xp.log(-np.pi / self.negative_half_precision))
        return xp.sum(components, axis=-1)

    @override
    def to_exp(self) -> MultivariateDiagonalNormalEP:
        xp = self.get_namespace()
        mean = -self.mean_times_precision / (2.0 * self.negative_half_precision)
        second_moment = xp.square(mean) - 0.5 / self.negative_half_precision
        return MultivariateDiagonalNormalEP(mean, second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> MultivariateDiagonalNormalEP:
        xp = get_namespace(x)
        return MultivariateDiagonalNormalEP(x, xp.square(x))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]

    def to_variance_parametrization(self) -> MultivariateDiagonalNormalVP:
        return self.to_exp().to_variance_parametrization()


@dataclass
class MultivariateDiagonalNormalEP(HasEntropyEP[MultivariateDiagonalNormalNP],
                                   ExpectationParametrization[MultivariateDiagonalNormalNP],
                                   Multidimensional,
                                   Samplable):
    """The expectation parametrization of the normal distribution with diagonal variance.

    Args:
        mean: E(x).
        second moment: The diagonal elements of E(x^2).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    second_moment: JaxRealArray = distribution_parameter(VectorSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[MultivariateDiagonalNormalNP]:
        return MultivariateDiagonalNormalNP

    @override
    def to_nat(self) -> MultivariateDiagonalNormalNP:
        return MultivariateDiagonalNormalNP(self.mean / self.variance(), -0.5 / self.variance())

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_variance_parametrization().sample(key, shape)

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> JaxRealArray:
        xp = self.get_namespace()
        return self.second_moment - xp.square(self.mean)

    def to_variance_parametrization(self) -> MultivariateDiagonalNormalVP:
        return MultivariateDiagonalNormalVP(self.mean, self.variance())


@dataclass
class MultivariateDiagonalNormalVP(Samplable, Multidimensional):
    """The variance parametrization of the normal distribution with diagonal variance.

    Args:
        mean: E(x).
        variance: The diagonal elements of Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    variance: JaxRealArray = distribution_parameter(VectorSupport())

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        xp = self.get_namespace()
        deviation = xp.sqrt(self.variance)
        return jax.random.normal(key, shape) * deviation + self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def pdf(self, x: JaxRealArray) -> JaxRealArray:
        return self.to_nat().pdf(x)

    def to_exp(self) -> MultivariateDiagonalNormalEP:
        xp = self.get_namespace()
        second_moment = self.variance + xp.square(self.mean)
        return MultivariateDiagonalNormalEP(self.mean, second_moment)

    def to_nat(self) -> MultivariateDiagonalNormalNP:
        precision = 1.0 / self.variance
        mean_times_precision = self.mean * precision
        negative_half_precision = -0.5 * precision
        return MultivariateDiagonalNormalNP(mean_times_precision, negative_half_precision)



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Protocol, override

from array_api_compat import get_namespace
from jax import Array


class _IndexUpdateHelper(Protocol):
    def __getitem__(self, index: Any) -> _IndexUpdateRef: ...


class _IndexUpdateRef(Protocol):
    def set(self, values: Any) -> Array: ...


@dataclass
class GenericIndexUpdateRef(_IndexUpdateRef):
    x: Any
    index: Any

    @override
    def set(self, values: Any) -> Array:
        xp = get_namespace(self.x)
        retval = xp.asarray(self.x, copy=True)
        retval[self.index] = values
        return retval


@dataclass
class GenericIndexUpdateHelper(_IndexUpdateHelper):
    x: Any

    @override
    def __getitem__(self, index: Any) -> GenericIndexUpdateRef:
        return GenericIndexUpdateRef(self.x, index)


def array_at(x: Any) -> _IndexUpdateHelper:
    if isinstance(x, Array):
        return x.at
    return GenericIndexUpdateHelper(x)



from __future__ import annotations

from typing import Any

from array_api_compat import get_namespace
from jax.random import split
from tjax import JaxIntegralArray, JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.samplable import Samplable
from ..natural_parametrization import NaturalParametrization
from ..parameter import (IntegralRing, ScalarSupport, distribution_parameter, negative_support,
                         positive_support)
from .gamma import GammaNP
from .negative_binomial_common import NBCommonEP, NBCommonNP
from .poisson import PoissonEP


@dataclass
class NegativeBinomialNP(Samplable,
                         NBCommonNP['NegativeBinomialEP'],
                         NaturalParametrization['NegativeBinomialEP', JaxRealArray]):
    """The natural parametrization of the negative binomial distribution.

    Models the number of Bernoulli trials having probability p until r failures.

    Args:
        log_not_p: log(1-p).
        failures: r.
    """
    log_not_p: JaxRealArray = distribution_parameter(ScalarSupport(ring=negative_support))
    failures: JaxIntegralArray = distribution_parameter(ScalarSupport(ring=IntegralRing(minimum=1)),
                                                        fixed=True)

    @override
    def to_exp(self) -> NegativeBinomialEP:
        return NegativeBinomialEP(self._mean(), self.failures)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> NegativeBinomialEP:
        xp = get_namespace(x)
        return NegativeBinomialEP(x, xp.broadcast_to(fixed_parameters['failures'], x.shape))

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)

    @override
    def _failures(self) -> JaxIntegralArray:
        return self.failures


@dataclass
class NegativeBinomialEP(Samplable,
                         NBCommonEP[NegativeBinomialNP],
                         ExpectationParametrization[NegativeBinomialNP]):
    """The expectation parametrization of the negative binomial distribution.

    Models the number of Bernoulli trials x having probability p until r failures.

    Args:
        mean: E(x).
        failures: r.
    """
    mean: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))
    failures: JaxIntegralArray = distribution_parameter(ScalarSupport(ring=IntegralRing(minimum=1)),
                                                        fixed=True)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[NegativeBinomialNP]:
        return NegativeBinomialNP

    @override
    def to_nat(self) -> NegativeBinomialNP:
        return NegativeBinomialNP(self._log_not_p(), self.failures)

    @override
    def _failures(self) -> JaxIntegralArray:
        return self.failures

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        gamma_key, poisson_key = split(key)
        r = self._failures()
        negative_rate = -r / self.mean
        shape_minus_one = r - 1.0
        g = GammaNP(negative_rate, shape_minus_one)
        lambda_ = g.sample(gamma_key, shape)
        p = PoissonEP(lambda_)
        return p.sample(poisson_key)



"""The generalized Dirichlet distribution.

It is based off:

T.-T. Wong 1998. Generalized Dirichlet distribution in Bayesian analysis. Applied Mathematics
and Computation, volume 97, pp165-181
"""
from __future__ import annotations

from typing import Any

from array_api_compat import get_namespace
from jax.scipy.special import betaln, digamma
from tjax import JaxRealArray, Shape, softplus
from tjax.dataclasses import dataclass
from typing_extensions import override

from ..expectation_parametrization import ExpectationParametrization
from ..interfaces.multidimensional import Multidimensional
from ..mixins.exp_to_nat.exp_to_nat import ExpToNat
from ..mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ..natural_parametrization import NaturalParametrization
from ..parameter import RealField, VectorSupport, distribution_parameter, negative_support


@dataclass
class GeneralizedDirichletNP(HasEntropyNP['GeneralizedDirichletEP'],
                             NaturalParametrization['GeneralizedDirichletEP', JaxRealArray],
                             Multidimensional):
    alpha_minus_one: JaxRealArray = distribution_parameter(VectorSupport(
        ring=RealField(minimum=-1.0, generation_scale=3.0)))
    gamma: JaxRealArray = distribution_parameter(VectorSupport(
        ring=RealField(minimum=0.0, generation_scale=3.0)))

    @property
    @override
    def shape(self) -> Shape:
        return self.alpha_minus_one.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        alpha, beta = self.alpha_beta()
        return xp.sum(betaln(alpha, beta), axis=-1)

    @override
    def to_exp(self) -> GeneralizedDirichletEP:
        # Given a log-normalizer y.
        # alpha_bar_direct = d y / d alpha = betaln'(alpha, beta) = digamma(alpha) - digamma(alpha +
        # beta)
        # beta_bar = d y / d beta = betaln'(alpha, beta) = digamma(beta) - digamma(alpha + beta)
        xp = self.get_namespace()
        alpha, beta = self.alpha_beta()
        digamma_sum = digamma(alpha + beta)
        alpha_bar_direct = digamma(alpha) - digamma_sum
        beta_bar = digamma(beta) - digamma_sum
        cs_beta_bar = xp.cumulative_sum(beta_bar, axis=-1, include_initial=True)
        gamma_bar = cs_beta_bar[..., 1:]
        alpha_bar_indirect = cs_beta_bar[..., :-1]
        alpha_bar = alpha_bar_direct + alpha_bar_indirect
        return GeneralizedDirichletEP(alpha_bar, gamma_bar)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> GeneralizedDirichletEP:
        xp = get_namespace(x)
        cs_x = xp.cumulative_sum(x, axis=-1)
        # cs_x[i] = sum_{j<=i} x[j]
        return GeneralizedDirichletEP(xp.log(x), xp.log1p(-cs_x))

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = self.get_namespace(x)
        return xp.zeros(x.shape[:len(x.shape) - 1])

    @override
    def dimensions(self) -> int:
        return self.alpha_minus_one.shape[-1]

    def alpha_beta(self) -> tuple[JaxRealArray, JaxRealArray]:
        xp = self.get_namespace()
        alpha = self.alpha_minus_one + 1.0
        # cs_alpha[i] = sum_{j>=i} alpha[j]
        # cs_gamma[i] = sum_{j>=i} gamma[j]
        cs_alpha = xp.cumulative_sum(alpha[..., ::-1], axis=-1, include_initial=True)[..., :-1]
        cs_gamma = xp.cumulative_sum(self.gamma[..., ::-1], axis=-1)
        beta = (cs_gamma + cs_alpha + 1.0)[..., ::-1]
        return alpha, beta


@dataclass
class GeneralizedDirichletEP(HasEntropyEP[GeneralizedDirichletNP],
                             ExpToNat[GeneralizedDirichletNP],
                             ExpectationParametrization[GeneralizedDirichletNP],
                             Multidimensional):
    # E({log(x_i)}_i)
    mean_log_probability: JaxRealArray = distribution_parameter(VectorSupport(
        ring=negative_support))
    # E({log(1-∑_{j≤i} x_j)}_i)
    mean_log_cumulative_probability: JaxRealArray = distribution_parameter(VectorSupport(
        ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean_log_probability.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[GeneralizedDirichletNP]:
        return GeneralizedDirichletNP

    @override
    def search_to_natural(self, search_parameters: JaxRealArray) -> GeneralizedDirichletNP:
        # Run Newton's method on the whole real hyperspace.
        n = self.dimensions()
        positive_search_parameters = softplus(search_parameters)
        return GeneralizedDirichletNP(positive_search_parameters[..., :n] - 1.0,
                                      positive_search_parameters[..., n:])

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def dimensions(self) -> int:
        return self.mean_log_probability.shape[-1]



"""Bayesian evidence combination.

This example is based on section 1.2.1 from expfam.pdf, entitled Bayesian evidence combination.

Suppose you have a prior, and a set of likelihoods, and you want to combine all of the evidence
into one distribution.
"""
from operator import add

import jax.numpy as jnp
from tjax import print_generic

from efax import MultivariateDiagonalNormalVP, parameter_map

prior = MultivariateDiagonalNormalVP(mean=jnp.zeros(2),
                                     variance=10 * jnp.ones(2))
likelihood = MultivariateDiagonalNormalVP(mean=jnp.asarray([1.1, -2.2]),
                                          variance=jnp.asarray([3.0, 1.0]))

# Convert to the natural parametrization.
prior_np = prior.to_nat()
likelihood_np = likelihood.to_nat()

# Sum.  We use parameter_map to ensure that we don't accidentally add "fixed" parameters, e.g., the
# failure count of a negative binomial distribution.
posterior_np = parameter_map(add, prior_np, likelihood_np)

# Convert to the source parametrization.
posterior = posterior_np.to_variance_parametrization()
print_generic(posterior)
# MultivariateDiagonalNormalVP[dataclass]
# ├── mean=Jax Array (2,) float32
# │   └──  0.0355 │ -0.2000
# └── variance=Jax Array (2,) float32
#     └──  0.0968 │ 0.0909



from __future__ import annotations

import math
from typing import Any

import jax
from tjax import JaxComplexArray, JaxRealArray, KeyArray, Shape, abs_square
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import ScalarSupport, complex_field, distribution_parameter


@dataclass
class ComplexUnitNormalNP(HasEntropyNP['ComplexUnitNormalEP'],
                          NaturalParametrization['ComplexUnitNormalEP', JaxComplexArray],
                          Samplable):
    """The natural parametrization of the complex unit normal distribution.

    It has unit variance, and zero pseudo-variance. This is a curved exponential family.

    Args:
        two_mean_conjugate: 2 * E(conjugate(x)).
    """
    two_mean_conjugate: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))
    # S = I, U = 0
    # P = I, R = 0
    # H = -I, J = 0
    # K = 0, L = I/2
    # eta = 2mean.conjugate()
    # Leta = mean.conjugate()

    @property
    @override
    def shape(self) -> Shape:
        return self.two_mean_conjugate.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @override
    def log_normalizer(self) -> JaxRealArray:
        mean_conjugate = self.two_mean_conjugate * 0.5
        return abs_square(mean_conjugate) + math.log(math.pi)

    @override
    def to_exp(self) -> ComplexUnitNormalEP:
        xp = self.get_namespace()
        return ComplexUnitNormalEP(xp.conj(self.two_mean_conjugate) * 0.5)

    @override
    def carrier_measure(self, x: JaxComplexArray) -> JaxRealArray:
        return -abs_square(x)

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxComplexArray, **fixed_parameters: Any
                              ) -> ComplexUnitNormalEP:
        return ComplexUnitNormalEP(x)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        return self.to_exp().sample(key, shape)


@dataclass
class ComplexUnitNormalEP(HasEntropyEP[ComplexUnitNormalNP],
                          ExpectationParametrization[ComplexUnitNormalNP],
                          Samplable):
    """The expectation parametrization of the complex unit normal distribution.

    It has unit variance, and zero pseudo-variance. This is a curved exponential family.

    Args:
        mean: E(x).
    """
    mean: JaxComplexArray = distribution_parameter(ScalarSupport(ring=complex_field))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape

    @override
    @classmethod
    def domain_support(cls) -> ScalarSupport:
        return ScalarSupport(ring=complex_field)

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[ComplexUnitNormalNP]:
        return ComplexUnitNormalNP

    @override
    def to_nat(self) -> ComplexUnitNormalNP:
        xp = self.get_namespace()
        return ComplexUnitNormalNP(xp.conj(self.mean) * 2.0)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        # The second moment of a normal distribution with the given mean.
        return -(abs_square(self.mean) + 1.0)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxComplexArray:
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        a = jax.random.normal(key, shape)
        b = jax.random.normal(key, shape)
        return a + 1j * b + self.mean

    # def conjugate_prior_distribution(self, n: JaxRealArray) -> IsotropicNormalNP:
    #     negative_half_precision = -0.5 * n * xp.ones(self.shape)
    #     return IsotropicNormalNP(n * self.mean, negative_half_precision)
    #
    # def conjugate_prior_observation(self) -> JaxComplexArray:
    #     return self.mean



from typing import Any

import optimistix as optx
from jax import jit
from tjax import JaxRealArray
from tjax.dataclasses import dataclass, field
from typing_extensions import override

from .exp_to_nat import ExpToNat, ExpToNatMinimizer

type RootFinder[Y, Out, Aux] = (
        optx.AbstractRootFinder[Y, Out, Aux, Any]
        | optx.AbstractLeastSquaresSolver[Y, Out, Aux, Any]
        | optx.AbstractMinimiser[Y, Aux, Any])


@dataclass
class OptimistixRootFinder(ExpToNatMinimizer):
    solver: RootFinder[JaxRealArray, JaxRealArray, Any] = field(static=True)
    max_steps: int = field(static=True)
    send_lower_and_upper: bool = field(static=True, default=False)

    @override
    def solve(self, exp_to_nat: ExpToNat[Any]) -> JaxRealArray:
        xp = exp_to_nat.get_namespace()

        @jit
        def f(x: JaxRealArray, args: None, /) -> JaxRealArray:
            if self.send_lower_and_upper:
                x = x[xp.newaxis]
            retval = exp_to_nat.search_gradient(x)
            if self.send_lower_and_upper:
                retval = retval[0]
            return retval

        initial = exp_to_nat.initial_search_parameters()
        if self.send_lower_and_upper:
            options = {}
            assert initial.shape == (1,)
            initial = initial[0]
            options['lower'] = initial - 1.0
            options['upper'] = initial + 1.0
        else:
            options = None
        results: optx.Solution[JaxRealArray, None] = optx.root_find(
                f, self.solver, initial, max_steps=self.max_steps, throw=False, options=options)
        if self.send_lower_and_upper:
            return results.value[xp.newaxis]
        return results.value


default_minimizer = OptimistixRootFinder(
        solver=optx.Newton[JaxRealArray, JaxRealArray, None](rtol=0.0, atol=1e-7),
        max_steps=1000)


default_bisection_minimizer = OptimistixRootFinder(
        solver=optx.Bisection(  # type: ignore[call-arg]
            rtol=0.0, atol=1e-7, flip='detect', expand_if_necessary=True),
        max_steps=1000,
        send_lower_and_upper=True)



from __future__ import annotations

from typing import Any, Generic, TypeVar

import numpy as np
import numpy.typing as npt
from numpy.random import Generator
from tjax import NumpyComplexArray, NumpyRealArray, Shape
from typing_extensions import override

T = TypeVar('T')


class ShapedDistribution(Generic[T]):
    """Allow a distributions with shape."""
    @override
    def __init__(self,
                 shape: Shape,
                 rvs_shape: Shape,
                 rvs_dtype: np.dtype[Any],
                 objects: npt.NDArray[np.object_]
                 ) -> None:
        super().__init__()
        self.shape = shape
        self.rvs_shape = rvs_shape
        self.rvs_dtype = rvs_dtype
        self.real_dtype: np.dtype[Any] = np.real(np.zeros(0, dtype=rvs_dtype)).dtype
        self.objects = objects

    def rvs(self,
            size: int | Shape | None = None,
            random_state: Generator | None = None) -> NumpyRealArray:
        if size is None:
            size = ()
        elif isinstance(size, int):
            size = (size,)
        retval = np.empty(self.shape + size + self.rvs_shape,
                          dtype=self.rvs_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].rvs(size=size, random_state=random_state)
        return retval

    def pdf(self, x: NumpyComplexArray) -> NumpyRealArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            value = self.objects[i].pdf(x[i])
            if i == ():
                return value
            retval[i] = value
        return retval

    def entropy(self) -> NumpyRealArray:
        retval = np.empty(self.shape, dtype=self.real_dtype)
        for i in np.ndindex(*self.shape):
            retval[i] = self.objects[i].entropy()
        return retval

    def access_object(self, index: tuple[int, ...]) -> T:
        return self.objects[index]



from __future__ import annotations

from typing import Any

import jax
from array_api_compat import get_namespace
from tjax import JaxRealArray, KeyArray, Shape
from tjax.dataclasses import dataclass
from typing_extensions import override

from ...expectation_parametrization import ExpectationParametrization
from ...interfaces.multidimensional import Multidimensional
from ...interfaces.samplable import Samplable
from ...mixins.has_entropy import HasEntropyEP, HasEntropyNP
from ...natural_parametrization import NaturalParametrization
from ...parameter import (ScalarSupport, VectorSupport, distribution_parameter, negative_support,
                          positive_support)


@dataclass
class IsotropicNormalNP(HasEntropyNP['IsotropicNormalEP'],
                        Samplable,
                        NaturalParametrization['IsotropicNormalEP', JaxRealArray],
                        Multidimensional):
    """The natural parametrization of the multivariate normal distribution with Var(x) = kI.

    Args:
        mean_times_precision: E(x) / Var(x).
        negative_half_precision: -0.5 / Var(x).
    """
    mean_times_precision: JaxRealArray = distribution_parameter(VectorSupport())
    negative_half_precision: JaxRealArray = distribution_parameter(ScalarSupport(
        ring=negative_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.negative_half_precision.shape

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @override
    def log_normalizer(self) -> JaxRealArray:
        xp = self.get_namespace()
        eta = self.mean_times_precision
        return 0.5 * (-0.5 * xp.sum(xp.square(eta), axis=-1) / self.negative_half_precision
                      + self.dimensions() * xp.log(xp.pi / -self.negative_half_precision))

    @override
    def to_exp(self) -> IsotropicNormalEP:
        xp = self.get_namespace()
        precision = -2.0 * self.negative_half_precision
        mean = self.mean_times_precision / precision[..., xp.newaxis]
        total_variance = self.dimensions() / precision
        total_second_moment = xp.sum(xp.square(mean), axis=-1) + total_variance
        return IsotropicNormalEP(mean, total_second_moment)

    @override
    def carrier_measure(self, x: JaxRealArray) -> JaxRealArray:
        xp = get_namespace(x)
        return xp.zeros(x.shape[:-1])

    @override
    @classmethod
    def sufficient_statistics(cls, x: JaxRealArray, **fixed_parameters: Any
                              ) -> IsotropicNormalEP:
        xp = get_namespace(x)
        return IsotropicNormalEP(x, xp.sum(xp.square(x), axis=-1))

    @override
    def dimensions(self) -> int:
        return self.mean_times_precision.shape[-1]

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        return self.to_exp().sample(key, shape)


@dataclass
class IsotropicNormalEP(HasEntropyEP[IsotropicNormalNP],
                        ExpectationParametrization[IsotropicNormalNP],
                        Samplable,
                        Multidimensional):
    """The expectation parametrization of the multivariate normal distribution with Var(x) = kI.

    Args:
        mean: E(x).
        variance: Var(x).
    """
    mean: JaxRealArray = distribution_parameter(VectorSupport())
    total_second_moment: JaxRealArray = distribution_parameter(ScalarSupport(ring=positive_support))

    @property
    @override
    def shape(self) -> Shape:
        return self.mean.shape[:-1]

    @override
    @classmethod
    def domain_support(cls) -> VectorSupport:
        return VectorSupport()

    @classmethod
    @override
    def natural_parametrization_cls(cls) -> type[IsotropicNormalNP]:
        return IsotropicNormalNP

    @override
    def to_nat(self) -> IsotropicNormalNP:
        xp = self.get_namespace()
        variance = self.variance()
        negative_half_precision = -0.5 / variance
        mean_times_precision = self.mean / variance[..., xp.newaxis]
        return IsotropicNormalNP(mean_times_precision, negative_half_precision)

    @override
    def expected_carrier_measure(self) -> JaxRealArray:
        xp = self.get_namespace()
        return xp.zeros(self.shape)

    @override
    def sample(self, key: KeyArray, shape: Shape | None = None) -> JaxRealArray:
        xp = self.get_namespace()
        if shape is not None:
            shape += self.mean.shape
        else:
            shape = self.mean.shape
        deviation = xp.sqrt(self.variance())
        return jax.random.normal(key, shape) * deviation[..., xp.newaxis] + self.mean

    @override
    def dimensions(self) -> int:
        return self.mean.shape[-1]

    def variance(self) -> JaxRealArray:
        xp = self.get_namespace()
        dimensions = self.dimensions()
        return (self.total_second_moment - xp.sum(xp.square(self.mean), axis=-1)) / dimensions
